/****************************************************************/
/* headers                                                      */
/****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "general.h"
#include "util.h"
#include "nvect.h"
#include "nmatrix.h"
#include "viz.h"
#include "bonds.h"

/****************************************************************/
/* defines  													*/
/****************************************************************/
#define NOT_FOUND		(-1)
#define SAFE(s)			((s) ? s : "NULL")
#define PRNT_BOOL(n)	((n) ? "TRUE" : "FALSE")

/****************************************************************/
/* typedefs                                                     */
/****************************************************************/
typedef struct param{
	char	*name;
	int		num_args;
	int		type;
	double	min;
	double	max;
	int		num_choices;
	char	*vals[10];
	int		codes[10];
	}PARAM;

/****************************************************************/
/* external function                                            */
/****************************************************************/
extern void set_required_colors( DB *db );

/****************************************************************/
/* function prototypes                                          */
/****************************************************************/
void load_config( void *data );
void save_config( void *data );
int read_config_args( DB *db, int argc, char **argv );
int read_config_file( DB *db, int argc, char **argv );
void usage( char *str );

/*
 * private functions
 */
static int parse_config( DB *db );
static int param_index( char *name );
static int valid_arg( char *arg, int index, int *code );
static void print_config( FILE *fp, DB *db );
static void print_codes( FILE *fp, int index, char *str, int mode );

/****************************************************************/
/* global variables                                             */
/****************************************************************/
PARAM param_list[] = { \
{"window_width",	 	1,	INTEGER, 	400,	1000,		0			},
{"window_height",	 	1,	INTEGER, 	400,	1000,		0			},
{"graph_width",		 	1,	INTEGER, 	100, 	500,		0			},
{"graph_height",	 	1,	INTEGER, 	100, 	500,		0			},
{"interact_mode",	 	1,	STRING,	   	0,		0, 			4,
										{"rotate","pan","zoom","select" },
										{ ROTATE,  PAN,  ZOOM,  SELECT  }},
{"output_mode",		 	2,	SPECIAL,   	0,		0, 			2,
													 {"text", "graphics"},
													 { TEXT,   GRAPHICS }},
{"draw_mode",		 	10, SPECIAL,   	0,		0, 			10, 
{ "dots", "lines", "trace", "labels", "bonds", "links",  "spheres", 
												"axes", "unit", "bounds" },
{  DOTS,   LINES,   TRACE,   LABELS,   ATTRACTIONS, LINKS, SPHERES, 
												 AXES,   UNIT,   BOUNDS  }},
{"plot_mode",		 	3,	SPECIAL,   	0,		0,			4,
{"none","stress","shepard","correlation"},
{ FALSE, STRESS, SHEPARD, CORRELATION }},
{"output_filename",	 	1, 	STRING,	   	0,		0, 			0			},
{"select_mode",		 	1, 	STRING,	 	0,		0,			2,
												  {"top_ten","random"},
												  { TOP_TEN,  RANDOM_ONCE }},
{"bonds_per_datum",	 	1, 	LONG, 		2, 		LONG_MAX, 	0		},
{"distance_function", 	1,	STRING,	   	0,		0,  		5, 
		{"euclidian", "manhattan", "cosine", "correlation", "minkowski" },
		{ EUCLIDIAN,   MANHATTAN,   COSINE,   CORRELATION,   MINKOWSKI }},
{"stress_function",	 	1,	STRING,	   	0,		0,			7, 
{ "kruskal1", "kruskal2", "lingoes", "sammon", "j_ee", "j_ff", "j_ef" },
{ KRUSKAL_STRESS1, KRUSKAL_STRESS2, LINGOES_STRESS, SAMMON_STRESS, 
  J_EE_STRESS, J_FF_STRESS, J_EF_STRESS }},
{"step_factor",		 	1,	FLOAT,	  	1.0,	3.0,		0			},
{"noise_factor",	 	1,	FLOAT,	  	0.0,	1.0,		0			},
{"gen_screen_coords",	1,	STRING,		0,		0,			3, 
											{"random", "variance", "data"},
											{ RANDOM,   VARIANCE,   DATA }},
{"distance_expression",	1,	STRING,		0,		0,			0			},
{"stress_expression",	1,	STRING,		0,		0,			0			},
{"display_dimensions",	1,	INTEGER,	1,	   	MAX_DISPLAY_DIMS, 0		},
{"viewplane_minimum",	2,	FLOAT,	  	0.0,	0.0,		0			},
{"viewplane_maximum",	2,	FLOAT,	  	0.0,	0.0,		0			},
{"eye_point",		 	3,	FLOAT,	  	0.0,	0.0,		0			},
{"lookat_point",	 	3, 	FLOAT,    	0.0,    0.0,		0			},
{"up_vector",		 	3, 	FLOAT,    	0.0,    0.0,		0			},
{"eye_to_viewplane", 	1, 	FLOAT,    	0.1,    40.0,		0			},	
{"eye_to_front_clip",	1,	FLOAT,	  	0.1,	40.0,		0			},
{"eye_to_back_clip", 	1, 	FLOAT,    	0.1,    40.0,		0			},
{"light_point",      	3, 	FLOAT,    	0.0,    0.0,		0			},
{"radius",				1,	FLOAT,	  	1.0,	40.0,		0			},
{"ambient_rgb",			3,  FLOAT,    	0.0,    1.0,		0			},
{"diffuse_rgb",			3,  FLOAT,    	0.0,    1.0,		0			},
{"specular_rgb",		3,  FLOAT,    	0.0,    1.0,		0			},
{"maximum_iteration",	1,	LONG,		0,	  	LONG_MAX, 	0			},
{"save_frames",			1,  STRING,		0,		0,			8,
		{ "yes", "no", "true", "false", "on", "off", "enabled", "disabled" },
		{  TRUE, FALSE, TRUE,   FALSE,  TRUE, FALSE,  TRUE,		FALSE   }},
{"frame_base_filename",	1,	STRING,		0,		0,			0			},
{"first_frame_saved",	1,	LONG,		0,	    LONG_MAX, 	0			},
{"last_frame_saved",	1,	LONG,		0,		LONG_MAX, 	0			},
{"skip_frames",			1,  INTEGER,	0,		LONG_MAX, 	0			},
{"data_source",			1,  STRING,		0,		0, 			8,
{"no_data","from_file","sim_tri","sim_cube","sim_hypercube","sim_cluster",
 "sim_tetra", "sim_tetrahedron" }, 
{ NO_DATA,  FROM_FILE,  SIM_TRI,  SIM_CUBE,  SIM_HYPERCUBE,  SIM_CLUSTER,
  SIM_TETRAHEDRON, SIM_TETRAHEDRON }}, 
{"data_filename",		1,	STRING,		0,		0,			0			},
{"normalize_function", 	1,	STRING,	   	0,		0,  		4, 
					{"none",   "scale", "subtract_mean", "unit_variance" },
					{ NO_NORM, SCALE,    SUB_MEAN,        UNIT_VARIANCE }},
{"interpolate",			1,  STRING,		0,		0,			8,
		{ "yes", "no", "true", "false", "on", "off", "enabled", "disabled" },
		{  TRUE, FALSE, TRUE,   FALSE,  TRUE, FALSE,  TRUE,		FALSE   }},
{ "num_interpolate_steps", 1, INTEGER, 0, 		10,			0			},
{ "back_color",			3,	INTEGER,  	0,		255,		0			},
{ "axis_color",			3,	INTEGER,  	0,		255,		0			},
{ "low_stress_color",	3,	INTEGER,  	0,		255,		0			},
{ "high_stress_color",	3,	INTEGER,  	0,		255,		0			},
{ "cluster_probability",1,  FLOAT,    	0.0,    1.0,		0			},
{ "num_data_points",	1,  LONG,    	1,    	LONG_MAX,	0			},
{ "num_data_dims",		1,  LONG,    	1,    	LONG_MAX,	0			},
{"",					0,	0,			0,		0,			0			},
};

char usage_msg[] = "Usage: mavis: [name=value ...]\n";
/*char usage_msg[] = \
"Usage: mavis: [name=value ...]\n\
\n\
where 'name=value' can be any of:\n\
\n\
Display options:\n\
----------------\n\
window_width=<num>              width of output window\n\
window_height=<num>             height of output window\n\
graph_width=<num>               width of graph windows\n\
graph_height=<num>              height of graph windows\n\
interact_mode=<mode>            [ ROTATE | PAN | ZOOM | SELECT]\n\
output_mode=<mode>              [ GRAPHICS | TEXT ]\n\
draw_mode=<mode>                [ LABELS | BONDS | SPHERES | DOTS |  \n\
                                  LINKS | TRACE ]\n\
plot_mode=<mode>                [ SHEPARD | STRESS | CORRELATION ]\n\
back_color=<num,num,num>        Background RGB color\n\
axis_color=<num,num,num>        Axis RGB color\n\
low_stress_color=<num,num,num>  Low Stress RGB\n\
high_stress_color=<num,num,num> High Stress RGB\n\
\n\
Data options:\n\
-------------\n\
\n\
data_source=<mode>              [ FROM_FILE | SIM_CLUSTER | SIM_CUBE | \n\
                                 SIM_TRI | SIM_TETRAHEDRON ]\n\
data_filename=<str>             input data filename\n\
output_filename=<str>           output data filename\n\
cluster_probability=<num>       clustering for simulated data\n\
num_data_points=<num>			number of simulated data points\n\
num_data_dims=<num>				number of simulated data dimensions\n\
\n\
MDS control parameters:\n\
-----------------------\n\
\n\
select_mode=<mode>              [ TOP_TEN | RANDOM ]\n\
bonds_per_datum=<num>           subset of data points to ``track''\n\
normalize_functio=<mode>        [ NO_NORM | SCALE | SUB_MEAN | \n\
                                  UNIT_VARIANCE ]\n\
distance_function=<mode>        [ EUCLIDIAN | MANHATTAN | COSINE | \n\
                                 CORRELATION | MINKOWSKI ]\n\
stress_function=<mode>          [ KRUSTAL1 | KRUSKAL2 | LINGOES | SAMMON |\n\
                                  EE | EF | FF ] \n\
step_factor=<num>               factor for scaling point updates\n\
noise_factor=<num>              factor for scaling additional noise\n\
gen_screen_coords=<mode>        [ RANDOM | VARIANCE ]\n\
distance_expression=<str>       user supplied distance calculation formula\n\
stress_expression=<str>         user supplied stress calculation formula\n\
display_dimensions=<num>        number of display (solution) dimensions\n\
\n\
view control parameters:\n\
------------------------\n\
\n\
viewplane_minimum=<num,num>     bottom left coordinates of vieplane\n\
viewplane_maximum=<num,num>     top right coordinates of vieplane\n\
eye_poin=<num,num,num>          x, y, z eye position in world coordinates\n\
lookat_point=<num,num,num>      x, y, z look at point in world coordinates\n\
up_vector=<num,num,num>         u, v, w vector of upward direction\n\
eye_to_viewplane=<num>          distance from eye to viewplane\n\
eye_to_front_clip=<num>         distance from eye to front clip plane\n\
eye_to_back_clip=<num>          distance from eye to back clip plane\n\
light_point=<num,num,num>       x, y, z light position world coordinates\n\
\n\
sphere control parameters:\n\
--------------------------\n\
\n\
radius=<num>                    radius of Z-buffered spheres\n\
ambient_rgb=<num,num,num>       Spheres' ambient RGB coefficients\n\
diffuse_rgb=<num,num,num>       Spheres' diffuse RGB coefficients\n\
specular_rgb=<num,num,num>      Spheres' specular RGB coefficients\n\
\n\
animation control parameters:\n\
-----------------------------\n\
\n\
maximum_iteration=<num>         number of frames in animation\n\
save_frames=<boolean>           save frames to disk (ON/OFF)\n\
frame_base_filename=<str>       default base filename for frames\n\
first_frame_saved=<num>         first frame to save\n\
last_frame_saved=<num>          last frame to save\n\
skip_frames=<num>               skip every <num>th frames\n\
interpolate=<boolean>           use Catmull-Rom iterpolation (ON/OFF)\n\
num_interpolate_steps=<num>     fineness of resolution in iterpolation\n";*/

char buf[1024];

int read_config_args( DB *db, int argc, char **argv )
{
	int		stat;

	DEBUG( "read_config_args" );

	reset_tokens( TOKEN_CMDLINE, NULL, argc, argv );
	stat = parse_config( db );

	return( stat );
}

int read_config_file( DB *db, int argc, char **argv )
{
	FILE	*fp;
	int		stat;

	DEBUG( "read_config_file" );

	if( (fp = fopen("mds.cfg", "r")) == NULL )
		return( ERROR );

	reset_tokens( TOKEN_FILE, fp, 0, NULL );
	stat = parse_config( db );

	fclose( fp );

	return( stat );
}

void load_config( void *data )
{
    DB      *db = (DB *)data;
    char    *filename;
	FILE	*fp;
	int		stat;
	long	k = db->mds.num_bonds;
	long	dims = db->dat.display_dims;

	DEBUG( "load_config" );

	filename = GetString("\nEnter filename for config file", "mds.cfg");
    if( filename == NULL )
		return;

	if( (fp = fopen(filename, "r")) == NULL )
	{
		print_error( "Error opening config file" );
		return;
	}
		
	reset_tokens( TOKEN_FILE, fp, 0, NULL );
	if( (stat = parse_config( db )) == ERROR )
		print_error( "Error reading config file" );

	fclose( fp );

	/*
     * if the display dimensionality has changed
     * free existing display arrays and allocate new ones!
     */
	if( dims != db->dat.display_dims && allocate_display_matrices(db) == ERROR )
    {
		print_error( "Error allocating display matrices" );
		return;
    }
	
	if( k != db->mds.num_bonds && allocate_subset_matrices(db) == ERROR )
    {
		print_error( "Error allocating subset matrices" );
		return;
    }

	/* reset sphere cache here - if needed */

	/*
     * restore un-normalized data to data patrix
     */
    copy_matrix( db->dat.A, db->dat.R );
    normalize_matrix( db->dat.A );

    /*
     * now that we have data - calculate
     * N-dimensional distances - ONE TIME ONLY :-)
     */
    calc_all_distances( db );

	/*
     * store shades of red to blue in private color cells
     */
	set_required_colors( db );
    calc_stress_colors( db );

	set_mode_toggles( db->draw_mode );

    draw_db( db );
}

void save_config( void *data )
{
    DB      *db = (DB *)data;
    char    *filename;
	FILE	*fp;
	int		stat;

	DEBUG( "save_config" );

	filename = GetString("\nEnter filename for config file", "mds.cfg");
    if( filename == NULL )
		return;

	if( (fp = fopen(filename, "w")) == NULL )
	{
		print_error( "Error opening config file for writing" );
		return;
	}
		
	print_config( fp, db );

	fclose( fp );
}

int parse_config( DB *db )
{
	char	tok[200];
	char	eq[10];
	char	val[10][200];
	int		i, index;
	int		code[10];
	double	dval[10];
	double	min, max;
	char	*end;
	char	*ptr;
	int		output_mode = 0;
	int		draw_mode = 0;
	int		plot_mode = 0;
	int		num_special_args;

	DEBUG( "parse_config" );

	while( get_token(tok,200) )
	{
		/*
		 * if parameter is specified without spaces (i.e. name=value)
		 * split parameter into parts and put '=' and "value" onto
		 * input buffer so they will be read off by next get_token()
		 */
		if( (ptr = strchr(tok,'=')) != NULL )
		{
			
			*ptr = '\0';
			unget_token( ++ptr );
			unget_token( "=" );
		}

		/*
		 * see if token is valid parameter
		 */
		if( (index = param_index(tok)) == NOT_FOUND )
		{
			sprintf( buf, "Unrecognized program parameter: %s", tok );
			print_error( buf );
			return( ERROR );
		}

		/*
		 * check for required '=' sign
		 */
		if( !get_token(eq,10) || strcmp( eq, "=") != 0 )
		{
			sprintf( buf, "missing '=' in config parameter: %s", tok );
			print_error( buf );
			return( ERROR );
		}

		/*
		 * read in number of arguments appropriate for this parameter
		 */
		for( i = 0; i < param_list[index].num_args; i++ )
		{
			if( !get_token(val[i], 200) )
			{
				if( i > 0 && param_list[index].type == SPECIAL )
					break;
				else
				{
					sprintf(buf,"missing argument for config parameter:%s",tok);
					print_error( buf );
					return( ERROR );
				}
			}

			/*
			 * if the parameter has a comma in it, put
			 * whatever's after the comma back on the
			 * input stream. If we have read in a comma
			 * on its own just discard it and reread the
			 * next argument
			 */
			if( (ptr = strchr(val[i],',')) != NULL )
			{
				*ptr = '\0';

				if( strlen(ptr+1) > 0 )
					unget_token( ++ptr );

				if( strlen(val[i]) == 0 )
				{
					i--;
					continue;
				}
			}

			/*
			 * the "special" parameters can have a variable
			 * number of arguments, so we read parameters
			 * until we get an illegal one, and then put
			 * that back on the input stream
			 */
			if( param_list[index].type == SPECIAL && 
							!valid_arg(val[i],index,&code[i]) )
			{
				unget_token( val[i] );
				break;
			}
		}

		if( param_list[index].type == SPECIAL )
			num_special_args = i;

		/*
		 * if parameter is a string check it against
		 * the list of legal string entries
		 */
		if( param_list[index].type == STRING && 
			!valid_arg(val[0], index, &code[0]) )
		{
			sprintf(buf,"invalid argument parameter for %s: %s",tok,val[0]);
			print_error( buf );
			return( ERROR );
		}

		/*
		 * if the parameter is a number convert it to a double
		 * and if the table specifies a max and a min, clamp
		 * value between max and min
		 */
		if( param_list[index].type == INTEGER ||
			param_list[index].type == LONG ||
			param_list[index].type == FLOAT )
		{
			min = param_list[index].min;
			max = param_list[index].max;

			for( i = 0; i < param_list[index].num_args; i++ )
			{
				/*
				 * get value in double format
				 */
				dval[i] = strtod( val[i], &end );

				/*
				 * if min and max are not both set to zero, CLAMP
				 * values between min and max
				 */
				if(	!(FLT_ZERO(min) && FLT_ZERO(max)) )
					dval[i] = CLAMP( dval[i], min, max);
			}
		}

		/*
		 * store parameter in appropriate 
		 * variable(s) in database
		 */
		switch( index )
		{
			case 0:	
					db->view.params[SCREEN_MAX][X] = (int)dval[0];
					break;
			case 1:	
					db->view.params[SCREEN_MAX][Y] = (int)dval[0];
					break;
			case 2:
					db->shepard.width = (int)dval[0];
					db->stress.width = (int)dval[0];
					db->correlate.width = (int)dval[0];
					break;
			case 3:
					db->shepard.height = (int)dval[0];
					db->stress.height = (int)dval[0];
					db->correlate.height = (int)dval[0];
					break;
			case 4:
					db->interact_mode = code[0];
					break;
			case 5:
					for( i = 0; i < num_special_args; i++ )
						output_mode |= code[i];
					break;	
			case 6:
					for( i = 0; i < num_special_args; i++ )
						draw_mode |= code[i];
					break;	
			case 7:
					for( i = 0; i < num_special_args; i++ )
						plot_mode |= code[i];
					break;	
			case 8:
					if( strcmp(strip_quotes(val[0]), "NULL") != 0 )
						new_str( &db->files.outfile, val[0] );
					break;	
			case 9:
					db->mds.select_mode = code[0];
					break;	
			case 10:
					db->mds.num_bonds = (long)dval[0];
					db->mds.num_bonds += ODD(db->mds.num_bonds) ? 1: 0;
					break;	
			case 11:
					db->mds.dist_func = code[0];
					break;
			case 12:
					db->mds.stress_function = code[0];
					break;
			case 13:
					db->mds.step_factor = dval[0];
					break;
			case 14:
					db->mds.noise_factor = dval[0];
					break;
			case 15:
					db->mds.display_init_mode = code[0];
					break;
			case 16:
					if( strcmp(strip_quotes(val[0]), "NULL") != 0 )
						new_str( &db->mds.dist_expression, val[0]);
					break;
			case 17:
					if( strcmp(strip_quotes(val[0]), "NULL") != 0 )
						new_str( &db->mds.stress_expression, val[0]);
					break;
			case 18:
					db->dat.display_dims = (int)dval[0];
					break;
			case 19:
					set_vect( db->view.params[VIEW_MIN],dval[0],dval[1],0,1);
					break;
			case 20:
					set_vect( db->view.params[VIEW_MAX],dval[0],dval[1],0,1);
					break;
			case 21:
					set_vect( db->view.params[EYE], dval[0],dval[1],dval[2],1);
					copy_vect( db->view.params[OLD_EYE], db->view.params[EYE] );
					break;
			case 22:
				   set_vect( db->view.params[LOOK_AT],
							 dval[0], dval[1], dval[2], 1 );
					break;
			case 23:
					set_vect(db->view.params[UP],dval[0],dval[1],dval[2],1);
					break;
			case 24:
					set_vect(db->view.params[VIEW_DIST], dval[0], 0, 0, 1 );
					break;
			case 25:
					set_vect(db->view.params[FRONT], dval[0], 0, 0, 1 );
					break;
			case 26:
					set_vect(db->view.params[BACK], dval[0], 0, 0, 1 );
					break;
			case 27:
				  set_vect( db->view.params[LIGHT_PT],
							dval[0], dval[1], dval[2],1);
					break;
			case 28:
					db->spheres.radius = dval[0];
					break;
			case 29:
					set_vect( db->spheres.properties.ambient,
							  dval[0], dval[1], dval[2], 1 );
					break;
			case 30:
					set_vect( db->spheres.properties.diffuse,
							  dval[0], dval[1], dval[2], 1 );
					break;
			case 31:
					set_vect( db->spheres.properties.specular,
								dval[0], dval[1], dval[2], 1 );
					break;
			case 32:
					db->animate.max_iteration = (long)dval[0];
					break;
			case 33:
					db->animate.save_frames = code[0];
					break;
			case 34:
					if( strcmp(strip_quotes(val[0]), "NULL") != 0 )
						new_str( &db->animate.frame_base, val[0] );
					break;
			case 35:
					db->animate.start_save_iter = (long)dval[0];
					break;
			case 36:
					db->animate.end_save_iter = (long)dval[0];
					break;
			case 37:
					db->animate.save_skip = (int)dval[0];
					break;
			case 38:
					db->dat.data_init_mode = code[0];
					break;
			case 39:
					if( strcmp(strip_quotes(val[0]), "NULL") != 0 )
						new_str( &db->files.infile, val[0] );
					break;	
			case 40:
					db->mds.norm_function = code[0];
					set_normalize( db->mds.norm_function );
					break;
			case 41:
					db->animate.interpolate = code[0];
					break;
			case 42:
					db->animate.interp_steps = (int)dval[0];
					break;
			case 43:
					init_private_color( db, BACK_COLOR, 
										dval[0], dval[1], dval[2] );
					break;

			case 44:
				   init_private_color( db, AXIS_COLOR,
									   dval[0], dval[1], dval[2] );
					break;
			case 45:
					init_private_color( db, LOW_STRESS_COLOR,
					 				    dval[0], dval[1], dval[2] );
					break;
			case 46:
					init_private_color( db, HIGH_STRESS_COLOR,
									    dval[0], dval[1], dval[2] );
					break;
			case 47:
					db->dat.clust_prob = dval[0];
					break;
			case 48:
					db->dat.num_points = dval[0];
					break;
			case 49:
					db->dat.data_dims = dval[0];
					break;
			default:
					break;
		}
	}

	if( output_mode != 0 )
		db->output_mode = output_mode;
	
	if( draw_mode != 0 )
		db->draw_mode = draw_mode;
	
	if( plot_mode != 0 )
		db->plot_mode = plot_mode;

	return( SUCCESS );
}

int param_index( char *name )
{
	char	low_name[200];
	int		i;

	DEBUG( "param_index" );

	strcpy( low_name, name );
	tolower_string( low_name );

	for( i = 0; param_list[i].name[0] != '\0'; i++ )
	{
		if( streq(low_name, param_list[i].name) )
			return( i );
	}

	return( NOT_FOUND );
}

int valid_arg( char *arg, int index, int *code )
{
	int		choices = param_list[index].num_choices;
	char	low_arg[200];
	int		i;

	DEBUG( "valid_arg" );

	if( choices == 0 )
		return( TRUE );

	strcpy( low_arg, arg );
	tolower_string( low_arg );

	for( i = 0; i < choices; i++ )
	{
		if( streq(low_arg, param_list[index].vals[i]) )
		{
			*code = param_list[index].codes[i];
			return( TRUE );
		}
	}

	return( FALSE );
}

void usage( char *str )
{
	int		c;

	DEBUG( "usage" );

	if( str )
		printf( "%s\n", str );

	printf( "See usage message? (y/n)" );
	c = getchar();
	if( c == 'y' || c == 'Y' )
		printf( "%s", usage_msg );
}
	
void print_config( FILE *fp, DB *db )
{
	int		r, g, b;
	char	tmp[100];

	DEBUG( "save_config" );

	fprintf( fp, "#\n#display parameters\n#\n" );
	fprintf(fp, "window_width\t\t= %d\n",(int)db->view.params[SCREEN_MAX][X]);
    fprintf(fp, "window_height\t\t= %d\n",(int)db->view.params[SCREEN_MAX][Y]);
    fprintf(fp, "graph_width\t\t\t= %d\n",db->shepard.width );
    fprintf(fp, "graph_height\t\t= %d\n",db->shepard.height );
	
	print_codes( fp, 4, "interact_mode\t",  db->interact_mode );
	print_codes( fp, 5, "output_mode\t\t", db->output_mode );
	print_codes( fp, 6, "draw_mode\t\t", db->draw_mode );
	print_codes( fp, 7, "plot_mode\t\t", db->plot_mode );

	get_private_color( BACK_COLOR, r, g, b );
    fprintf(fp, "back_color\t\t\t= %d %d %d\n", r, g, b );

	get_private_color( AXIS_COLOR, r, g, b );
    fprintf(fp, "axis_color\t\t\t= %d %d %d\n", r, g, b );

	get_private_color( LOW_STRESS_COLOR, r, g, b );
    fprintf(fp, "low_stress_color\t= %d %d %d\n", r, g, b );

	get_private_color( HIGH_STRESS_COLOR, r, g, b );
    fprintf(fp, "high_stress_color\t= %d %d %d\n", r, g, b );

    fprintf(fp, "\n#\n# files\n#\n" );

	strcpy( tmp, param_list[39].vals[db->dat.data_init_mode] );
	toupper_string( tmp );
	fprintf( fp, "data_source\t\t\t= %s\n", tmp );

	fprintf(fp, "data_filename\t\t= %s\n", SAFE(db->files.infile) );
    fprintf(fp, "output_filename\t\t= %s\n", SAFE(db->files.outfile) );
    fprintf(fp, "cluster_probability\t= %f\n", db->dat.clust_prob );

    fprintf(fp, "\n#\n# MDS control parameters\n#\n" );
    fprintf(fp, "bonds_per_datum\t\t= %d\n", db->mds.num_bonds );
	print_codes( fp, 9, "select_mode\t\t", db->mds.select_mode );

	strcpy( tmp, param_list[41].vals[db->mds.norm_function] );
	toupper_string( tmp );
	fprintf( fp, "normalize_function\t= %s\n", tmp );

	strcpy( tmp, param_list[11].vals[db->mds.dist_func] );
	toupper_string( tmp );
	fprintf( fp, "distance_function\t= %s\n", tmp );

	strcpy( tmp, param_list[12].vals[db->mds.stress_function] );
	toupper_string( tmp );
	fprintf( fp, "stress_function\t\t= %s\n", tmp );

    fprintf(fp, "step_factor\t\t\t= %f\n", db->mds.step_factor );
    fprintf(fp, "noise_factor\t\t= %f\n", db->mds.noise_factor );

	print_codes( fp, 15, "gen_screen_coords", db->mds.display_init_mode );
	fprintf(fp, "distance_expression\t= %s\n", SAFE(db->mds.dist_expression) );
	
	fprintf(fp, "stress_expression\t= %s\n", SAFE(db->mds.stress_expression) );
    fprintf(fp, "display_dimensions	= %d\n", db->dat.display_dims );

	fprintf(fp, "\n#\n# view control parameters\n#\n" );
    fprintf(fp, "viewplane_minimum	= %f %f\n", db->view.params[VIEW_MIN][X],
												db->view.params[VIEW_MIN][Y] );
    fprintf(fp, "viewplane_maximum	= %f %f\n", db->view.params[VIEW_MAX][X],
												db->view.params[VIEW_MAX][Y] );
    fprintf(fp, "eye_point			= %f %f %f\n", db->view.params[EYE][X],
												   db->view.params[EYE][Y],
												   db->view.params[EYE][Z] );
    fprintf(fp, "lookat_point		= %f %f %f\n", db->view.params[LOOK_AT][X],
												   db->view.params[LOOK_AT][Y],
												   db->view.params[LOOK_AT][Z]);
    fprintf(fp, "up_vector			= %f %f %f\n", db->view.params[UP][X],
												   db->view.params[UP][Y],
												   db->view.params[UP][Z] );
    fprintf(fp, "eye_to_viewplane	= %f\n", db->view.params[VIEW_DIST][X] ); 
    fprintf(fp, "eye_to_front_clip	= %f\n", db->view.params[FRONT][X] ); 
    fprintf(fp, "eye_to_back_clip	= %f\n", db->view.params[BACK][X] ); 
    fprintf(fp, "light_point\t\t\t= %f %f %f\n",db->view.params[LIGHT_PT][X],
												db->view.params[LIGHT_PT][Y],
												db->view.params[LIGHT_PT][Z] );

    fprintf(fp, "\n#\n# sphere control parameters\n#\n" );
    fprintf(fp, "ambient_rgb			= %f %f %f\n", 
											db->spheres.properties.ambient[0],
											db->spheres.properties.ambient[1],
											db->spheres.properties.ambient[2] );
    fprintf(fp, "diffuse_rgb			= %f %f %f\n", 
											db->spheres.properties.diffuse[0],
											db->spheres.properties.diffuse[1],
											db->spheres.properties.diffuse[2] );
    fprintf(fp, "specular_rgb		= %f %f %f\n", 
											db->spheres.properties.specular[0],
											db->spheres.properties.specular[1],
											db->spheres.properties.specular[2]);

    fprintf(fp, "\n#\n# animation control parameters\n#\n" );
    fprintf(fp, "maximum_iteration	= %d\n", db->animate.max_iteration );

	fprintf(fp, "save_frames\t\t\t= %s\n",PRNT_BOOL(db->animate.save_frames));
    fprintf(fp, "frame_base_filename\t= \"%s\"\n",SAFE(db->animate.frame_base));
    fprintf(fp, "first_frame_saved	= %ld\n", db->animate.start_save_iter );
    fprintf(fp, "last_frame_saved	= %ld\n", db->animate.end_save_iter );
    fprintf(fp, "skip_frames\t\t\t= %d\n", db->animate.save_skip );

	fprintf(fp, "interpolate\t\t\t= %s\n",PRNT_BOOL(db->animate.interpolate));
    fprintf(fp, "num_interpolate_steps = %d\n", db->animate.interp_steps );
}

void print_codes( FILE *fp, int index, char *str, int mode )
{
	int		i, limit = param_list[index].num_choices;
	int		flag = 0;
	char	tmp[100];

	DEBUG( "valid_arg" );

	fprintf( fp, "%s\t= ", str );

	for( i = 0; i < limit; i++ )
	{
		if( (mode & param_list[index].codes[i]) == param_list[index].codes[i] )
		{
			if( flag == 0 )
				flag = 1; 
			else
				fprintf( fp, "%c", ',' );

			strcpy( tmp, param_list[index].vals[i] );
			toupper_string( tmp );
			fprintf( fp, "%s", tmp );
		}
	}

	fprintf( fp, "%c", '\n' );
}
