#ifndef MATRIX_H
#define MATRIX_H

/****************************************************************/
/* headers														*/
/****************************************************************/
#include <math.h>

/****************************************************************/
/* defines														*/
/****************************************************************/
#define X		0
#define Y		1
#define Z		2
#define W		3

/*
 * matrix macros
 */
#define dot( a, b )       			( a[X]*b[X] + a[Y]*b[Y] + a[Z]*b[Z] )

#define set_vect( v, x, y, z, w )	{ v[X]=(double)(x); v[Y]=(double)(y); \
									  v[Z]=(double)(z); v[W]=(double)(w); }

#define cross( a, b, c ) {	(c)[X] = (a)[Y]*(b)[Z] - (a)[Z]*(b)[Y]; 	\
							(c)[Y] = (a)[Z]*(b)[X] - (a)[X]*(b)[Z]; 	\
							(c)[Z] = (a)[X]*(b)[Y] - (a)[Y]*(b)[X]; 	\
						}

#define	veclen( v )		(sqrt((v)[X]*(v)[X]+(v)[Y]*(v)[Y]+(v)[Z]*(v)[Z]))

#define norm( v )		{	double len = veclen(v);						\
							double inv_len;								\
							if( ! FLT_ZERO(len) ) 						\
							{											\
								inv_len = 1.0/len;						\
								v[X] *= inv_len;						\
								v[Y] *= inv_len;						\
								v[Z] *= inv_len;						\
							}											\
						}

#define homo( p )		{	double inv_w;								\
							if( ! FLT_ZERO(p[W]) )                      \
                            {        									\
								inv_w = 1.0/p[W];						\
								p[X] *= inv_w;							\
								p[Y] *= inv_w;							\
								p[Z] *= inv_w;							\
								p[W] = 1.0;								\
							}											\
						}

#define eq_vect( v1, v2 )	(( ABS(dot(v1,v2)-1.0) < EPSILON ) ? TRUE : FALSE )		

#define equal_vect( v1, v2 ) ( v1[X]==v2[X] &&  v1[Y]==v2[Y] && v1[Z]==v2[Z] )

#define copy_vect( v1, v2 )		{										\
									v1[X] = v2[X];						\
									v1[Y] = v2[Y];						\
									v1[Z] = v2[Z];						\
									v1[W] = v2[W];						\
								}

#define sub_vect( v1, v2, r ) {											\
									r[X] = v1[X] - v2[X];				\
									r[Y] = v1[Y] - v2[Y];				\
									r[Z] = v1[Z] - v2[Z];				\
								}

#define add_vect( v1, v2, r ) {											\
									r[X] = v1[X] + v2[X];				\
									r[Y] = v1[Y] + v2[Y];				\
									r[Z] = v1[Z] + v2[Z];				\
								}
#define mult_vect( f, v )		{										\
									v[X] = v[X] * f;					\
									v[Y] = v[Y] * f;					\
									v[Z] = v[Z] * f;					\
								}

#define div_vect( f, v )		{	double inv_f;						\
									if( ! FLT_ZERO(f) ) 				\
									{									\
										inv_f = 1.0/f;					\
										v[X] = v[X] * inv_f;			\
										v[Y] = v[Y] * inv_f;			\
										v[Z] = v[Z] * inv_f;			\
									}									\
								}

#define vect_mult_vect( v1, v2, r ) {									\
									r[X] = v1[X] * v2[X];				\
									r[Y] = v1[Y] * v2[Y];				\
									r[Z] = v1[Z] * v2[Z];				\
								}

#define vect_mult_vect_add( v1, v2, r ) {								\
									r[X] += v1[X] * v2[X];				\
									r[Y] += v1[Y] * v2[Y];				\
									r[Z] += v1[Z] * v2[Z];				\
								}

#define avg_vect( v1, v2, r ) 	{										\
									r[X] = (v1[X] + v2[X])*0.5;			\
									r[Y] = (v1[Y] + v2[Y])*0.5;			\
									r[Z] = (v1[Z] + v2[Z])*0.5;			\
								}

/****************************************************************/ 
/* typedefs														*/
/****************************************************************/
typedef double VECT[4];
typedef double MAT4X4[4][4];
typedef double PT2D[2];
typedef VECT PT3D;

/****************************************************************/ 
/* prototypes													*/
/****************************************************************/
extern void print_mat( MAT4X4, char * );
extern void print_vect( VECT, char * );
extern void copy_mat( MAT4X4, MAT4X4 );
extern void transpose_mat( MAT4X4, MAT4X4 );
extern void ident_mat( MAT4X4 );
extern void zero_mat( MAT4X4 );
extern void init_trans( MAT4X4, double, double, double );
extern void init_scale( MAT4X4, double, double, double );
extern void init_xrot( MAT4X4, double );
extern void init_yrot( MAT4X4, double );
extern void init_zrot( MAT4X4, double );
extern void trans_mat( MAT4X4, double, double, double );
extern void scale_mat( MAT4X4, double, double, double );
extern void xrot_mat( MAT4X4, double );
extern void yrot_mat( MAT4X4, double );
extern void zrot_mat( MAT4X4, double );
extern void rotate_axis( VECT, double, MAT4X4 );
extern void scalar_mat_mult( double k, MAT4X4 M );
extern double **mat_mult( MAT4X4, MAT4X4, MAT4X4 );
extern double *pt_mat_mult( VECT, MAT4X4, VECT );
extern double dist( VECT, VECT );
extern void load_mat_rows( MAT4X4, VECT, VECT, VECT, VECT );
extern void load_mat_cols( MAT4X4, VECT, VECT, VECT, VECT, long );
extern void trackball( int, int, int, int, int, int, MAT4X4 );
extern int intersect_sphere( VECT, VECT, VECT, double, VECT );

#endif

