/****************************************************************/
/* headers                                                      */
/****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "general.h"
#include "util.h"
#include "viz.h"
#include "nvect.h"
#include "nmatrix.h"
#include "db.h"
#include "cluster.h"
#include "bonds.h"
#include "stats.h"
#include "file.h"
#include "animate.h"
#include "cfg.h"

/****************************************************************/
/* defines                                                      */
/****************************************************************/
#define ADD_BOND_NEAR( i, j, ndist, off_dist, offj ) 						  \
offj = -1, add_bond_sorted( nD->data[(i)], O->ldata[(i)], 0, t, (j), (ndist), 	  \
			 &db->dat.data[(i)].num_bonds, &(off_dist), &(offj), is_nearer)

#define ADD_BOND_FAR( i, j, ndist, off_dist, offj ) 						  \
offj = -1, add_bond_sorted( nD->data[(i)], O->ldata[(i)], t, limit, (j), (ndist), \
			 &db->dat.data[(i)].num_bonds, &(off_dist), &(offj), is_further)

#define BOND_DROPPED( off )		((off) != -1)

/****************************************************************/
/* typedefs                                                     */
/****************************************************************/
typedef int (*TST_FUNC)( double, double );

/****************************************************************/
/* function prototypes                                          */
/****************************************************************/
void fill_all_distances( DB *db, double **nD );
void fill_screen_coords( DB *db, double **disP );
void get_result( DB *db, double **sD );

extern void setup_screen_coords( DB *db );
extern int add_bond_sorted( double dists[], long others[], long low,long high, 
						    long j, double ndist, long *num_bonds, 
							double *off_dist, long *offj, TST_FUNC tst );
extern int is_nearer( double new_dist, double old_dist );
extern int is_further( double new_dist, double old_dist );


int mds(int nP, int dim, double **dataD, int disDim, double **disP, int num_bonds, int num_iter, int useDisP)
{
	DB	*db = NULL;

	DEBUG( "main" );

	/*
	 * allocate and initialize database
	 */
	if( (db = alloc_db()) == NULL )
		return( ERROR );

	/*
	 * allocate data for database
	 */
	db->dat.num_points = nP;
	db->dat.data_dims = dim;
    db->dat.display_dims = disDim;
    db->mds.num_bonds = num_bonds;
    db->animate.max_iteration = num_iter;
	db->name = mstrdup( "MDS" );

	if( alloc_data( db ) == ERROR )
		return( ERROR );

    /*
	 * fill N-dimensional distances using data passed in
	 */
	fill_all_distances( db, dataD );

	/*
     * calculate screen coords
     */
    if (useDisP)
        fill_screen_coords( db, disP );
    else
        setup_screen_coords( db );

	/*
	 * update screen positions of datums
	 */
	animate_db( db );

	/*
	 * get result from database
	 */
	get_result( db, disP );

	/*
	 * clean up
	 */
	free_db( db );

    return( 0 );
}

/*
 * fill N-dimensional distances using data passed in
 */
void fill_all_distances( DB *db, double **dataD )
{
    long        num_points;
	long		offj, offj2;
	long		i, j, t, n = 0;
	double		off_dist, off_dist2;
	double		min = DBL_MAX;
	double		max = DBL_MIN;
	double		sum = 0.0;
	double		mean = 0.0;
	double		var = 0.0;
	double		ndist;

    MATRIX*     nD = db->dat.nD;
    MATRIX*     O = db->dat.O;
    long        limit = db->mds.num_bonds;
    int*        mask = db->dat.dim_mask;
    int         mode = db->mds.select_mode;
	MDS_STATS*  stats = &db->stats;

    DEBUG( "calc_all_ndist" );

	stats->num_dropped = 0;
	t = limit/2;

    num_points = db->dat.num_points;

	/*
	 * calculate all n(n-1)/2 distances between data points
	 * also calculate mean, min, max n-dimensional distances
	 */	
	for( i = 0; i < num_points; i++ )
	{
		db->dat.data[i].num_bonds = 0;

		for( j = 0; j < i; j++ )
		{
			ndist = dataD[i][j];

			sum += ndist;
			min = MIN( ndist, min );
			max = MAX( ndist, max );
			
			if( ADD_BOND_NEAR( i, j, ndist, off_dist, offj ) )
			{
				/*
				 * first try to add distance from i -> j to i's list
				 * see if it is a distance which belongs among the
				 * nearest k/2 distances. If it is added to the top set
				 * then it is possible that a distance was dropped off
				 * the list of the top k/2, so see if that bumped distance
				 * from i -> dropped data point can be added to the dropped 
				 * data point's near list
				 */
				DEBUG2( "Added %ld to %ld's near list ...\n", j, i );

				if( BOND_DROPPED(offj) )
				{
					if( ADD_BOND_NEAR( offj, i, off_dist, off_dist2, offj2 ) )
						DEBUG2( "Added %ld to %ld's near list ...\n", offj, i );
				}
			}
			else if( ADD_BOND_FAR( i, j, ndist, off_dist, offj ) )
			{
				/*
				 * next try to add distance from i -> j to i's list
				 * of far distances. If it is added and this forces
				 * a distance to be dropped, try to add the distance
				 * from i -> dropped data point to the dropped data
				 * points far list
			     */
				DEBUG2( "Added %ld to %ld's near list ...\n", i, j );

				if( BOND_DROPPED(offj) )
				{
					if( ADD_BOND_FAR( offj, i, off_dist, off_dist2, offj2 ) )
						DEBUG2( "Added %ld to %ld's far list ...\n", i, offj );
				}
			}
			else if( ADD_BOND_NEAR( j, i, ndist, off_dist, offj ) )
			{
				/*
				 * next try to add distance from i -> j to j's list
				 * of near distances. If it is added and this forces
				 * a distance to be dropped, try to add the distance
				 * from j -> dropped data point to the dropped data
				 * points far list
			     */
				DEBUG2( "Added %ld to %ld's near list ...\n", i, j );

				if( BOND_DROPPED(offj) )
				{
					if( ADD_BOND_FAR( j, offj, off_dist, off_dist2, offj2 ) )
						DEBUG2( "Added %ld to %ld's far list ...\n", offj, j );
				}
			}
			else if( ADD_BOND_FAR( j, i, ndist, off_dist2, offj2 ) )
			{
				/*
				 * next try to add distance from i -> j to j's list
				 * of far distances. If it doesn't go there - give up
				 * we tried hard enough :-)
			     */
				DEBUG2( "Added %ld to %ld's near list ...\n", i, j );
			}
			else
				db->stats.num_dropped++;
		}
	}
	
	/*
	 * divide by total to calc mean
	 */
	n = num_points * (num_points-1) / 2;
	mean = sum/n;
	
	/*
	 * store resulting stats in database struct
	 */
	stats->sum_ndist = sum;
	stats->num_ndist = n;
	stats->min_ndist = min;
	stats->max_ndist = max;
	stats->mean_ndist = mean;
}

/*
 * fill screen coords using data passed in
 */
void fill_screen_coords( DB *db, double **disP )
{
    MATRIX *P;
    int numP;
    int dim;
    int i, j;

	P = db->dat.P;
    numP = P->height;
    dim = P->width;
	for( i = 0; i < numP ; i++ )
	{
		for( j = 0; j < dim; j++ )
		{
			P->data[i][j] = disP[j][i];
		}
	}
}

/*
 * get result from database
 */
void get_result( DB *db, double **disP )
{
    MATRIX *P;
    int numP;
    int dim;
    int i, j;
    double min, max;
    double margin = 0.01;
    double rate;

	P = db->dat.P;
    numP = P->height;
    dim = P->width;
	for( i = 0; i < numP ; i++ )
	{
		for( j = 0; j < dim; j++ )
		{
			disP[j][i] =  P->data[i][j];
		}
	}

	for( i = 0; i < dim ; i++ )
	{
        min = max = disP[i][0];
		for( j = 1; j < numP; j++ )
		{
            min = MIN( min, disP[i][j] );
            max = MAX( max, disP[i][j] );
		}

        rate = (1 - 2 * margin) / (max - min);
		for( j = 0; j < numP; j++ )
        {
            disP[i][j] = (disP[i][j] - min) * rate + margin;
        }
	}
}