/****************************************************************/
/* headers                                                      */
/****************************************************************/
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <limits.h>
#include "general.h"
#include "util.h"
#include "nmatrix.h"
#include "nvect.h"

/****************************************************************/
/* defines                                                      */
/****************************************************************/
	
/****************************************************************/
/* typedefs                                                     */
/****************************************************************/

/****************************************************************/
/* external functions                                           */
/****************************************************************/
		
/****************************************************************/
/* function prototypes                                          */
/****************************************************************/
void set_normalize( int mode );
void set_global_stress( int mode );
MATRIX *alloc_matrix( long width, long height, int type );
void free_matrix_mem( MATRIX *A );
MATRIX *init_matrix( MATRIX *A, double val );
MATRIX *copy_matrix( MATRIX *dst, MATRIX *src );
MATRIX *dup_matrix( MATRIX *src );
void print_matrix( MATRIX *A, char *str );
MATRIX *matrix_matrix_mult( MATRIX *A, MATRIX *B );
MATRIX *scalar_matrix_mult( double k, MATRIX *A );
MATRIX *transpose_matrix( MATRIX *A );
MATRIX *matrix_matrix_add( MATRIX *A, MATRIX *B );
MATRIX *matrix_matrix_sub( MATRIX *A, MATRIX *B );
MATRIX *normalize_matrix( MATRIX *A );
MATRIX *matrix_transpose_mult( MATRIX *A );
MATRIX *correlation_matrix( MATRIX *A );
MATRIX *distance_matrix( MATRIX *A, int mask[] );
MATRIX *matrix_square( MATRIX *A );
double matrix_mean( MATRIX *A );
double matrix_summation( MATRIX *A );
double column_mean( MATRIX *A, long j );
double column_variance( MATRIX *A, long j );
double column_median( MATRIX *A, long j );
double column_min( MATRIX *A, long j );
double column_max( MATRIX *A, long j );
void column_noop( MATRIX *A, long j );
void column_scale( MATRIX *A, long j );
void column_unitvar( MATRIX *A, long j );
void column_submean( MATRIX *A, long j );
double calc_dim_correlation( MATRIX *A, long dim1, long dim2 );
double kruskal_stress( MATRIX *A, MATRIX *P, int mask[] );
double sammon_stress( MATRIX *A, MATRIX *P, int mask[] );
double lingoes_stress( MATRIX *A, MATRIX *P, int mask[] );
double J_ee_stress( MATRIX *A, MATRIX *P, int mask[] );
double J_ff_stress( MATRIX *A, MATRIX *P, int mask[] );
double J_ef_stress( MATRIX *A, MATRIX *P, int mask[] );
double dist_correlation( MATRIX *A, MATRIX *P, int mask[] );
double calc_max_distance( MATRIX *A, int mask[], long *index1, long *index2 );
double calc_min_distance( MATRIX *A, int mask[], long *index1, long *index2 );
double calc_mean_distance( MATRIX *A, int mask[] );
double max_distance_from_i( MATRIX *A, int mask[], long i, long *j_index );
double min_distance_from_i( MATRIX *A, int mask[], long i, long *j_index );

/****************************************************************/
/* global variables												*/
/****************************************************************/
void (*column_norm)( MATRIX *A, long j ) = column_unitvar;
double (*global_stress)( MATRIX *, MATRIX *, int mask[] ) = lingoes_stress;

/****************************************************************/
/*	function	set_normalize() - sets which data normalization	*/
/*				method to use when normalizing data matrix		*/
/*																*/
/*	input:		mode - integer id of normalization method		*/
/*																*/
/*	output:		none											*/
/****************************************************************/
void set_normalize( int mode )
{
	DEBUG( "set_normalize" );

	switch( mode )
	{
		case NO_NORM:
			column_norm = column_noop;
			break;

		case SCALE:
			column_norm = column_scale;
			break;

		case SUB_MEAN:
			column_norm = column_submean;
			break;
		
		case UNIT_VARIANCE:
		default:
			column_norm = column_unitvar;
            break;
	}
}

/****************************************************************/
/*  function    set_stress_func() - select flavor of STRESS to  */
/*              calculate for distance relations                */
/*                                                              */
/*  input:      db      - pointer to database                   */
/*                                                              */
/*  output:     none                                            */
/****************************************************************/
void set_global_stress( int mode )
{
	DEBUG( "set_global_stress" );

    switch( mode )
    {
        case J_EE_STRESS:
            global_stress = J_ee_stress;
            break;
        case J_FF_STRESS:
            global_stress = J_ff_stress;
            break;
        case J_EF_STRESS:
            global_stress = J_ef_stress;
            break;
        case SAMMON_STRESS:
            global_stress = sammon_stress;
            break;
        case KRUSKAL_STRESS1:
		case KRUSKAL_STRESS2:
            global_stress = kruskal_stress;
        case LINGOES_STRESS:
        default:
            global_stress = lingoes_stress;
            break;
    }
}

/****************************************************************/
/*	function	alloc_matrix() -allocate memory for arbitrary	*/
/*				m x n matrix of doubles							*/
/*																*/
/*	input:		width 	- long width of matrix					*/
/*				height	- long height of matrix					*/
/*				type	- int data type of matrix				*/
/*																*/
/*	output:		pointer to allocated matrix struct				*/
/****************************************************************/
MATRIX *alloc_matrix( long width, long height, int type )
{
	MATRIX 		*new;
	long		i, j;

	DEBUG( "alloc_matrix" );

	if( (new = (MATRIX *)mmalloc(sizeof(MATRIX))) == NULL )
		return( NULL );

	new->type = type;
	new->width = width;
	new->height = height;
	new->data = NULL;
	new->cdata = NULL;
	new->idata = NULL;
	new->sdata = NULL;
	new->ldata = NULL;
	new->fdata = NULL;

	switch( type )
	{
		case LONG_TYPE:
			if( !(new->ldata = (long **)mmalloc(height*sizeof(long *))) )
			{
				mfree( new );
				return( NULL );
			}

			for( i = 0; i < height; i++ )
			{
				if( !(new->ldata[i] = (long *)mmalloc(width*sizeof(long))) )
				{
					for( j = 0; j < i; j++ )
						mfree( new->ldata[j] );

					mfree( new );
					return( NULL );
				}
			}
			break;

		case DOUBLE_TYPE:
			if( !(new->data = (double **)mmalloc(height*sizeof(double *))) )
			{
				mfree( new );
				return( NULL );
			}

			for( i = 0; i < height; i++ )
			{
				if( !(new->data[i] = (double *)mmalloc(width*sizeof(double))) )
				{
					for( j = 0; j < i; j++ )
						mfree( new->data[j] );

					mfree( new );
					return( NULL );
				}
			}
			break;
	}

	return( new );
}

/****************************************************************/
/*	function	free_matrix_mem() - free memory used by matrix		*/
/*																*/
/*	input:		pointer to matrix to deallocate					*/
/*																*/
/*	output:		none											*/
/****************************************************************/
void free_matrix_mem( MATRIX *A )
{
	long		i;

	DEBUG( "free_matrix_mem" );

	if( !A )
		return;

	if( !A->data && !A->ldata )
	{
		mfree( A );
		return;
	}

	switch( A->type )
	{
		case LONG_TYPE:
				for( i = 0; i < A->height; i++ )
					mfree( A->ldata[i] );

				mfree( A->ldata );
				break;

		case DOUBLE_TYPE:
		default:
				for( i = 0; i < A->height; i++ )
					mfree( A->data[i] );

				mfree( A->data );
				break;
	}

	mfree( A );
}

/****************************************************************/
/*	function	init_matrix() - fill matrix with double value	*/
/*																*/
/*	input:		A	- pointer to matrix to fill					*/	
/*				val	- double value to initialize matrix to		*/
/*																*/
/*	output:		pointer to matrix								*/
/****************************************************************/
MATRIX *init_matrix( MATRIX *A, double val )
{
	long		i, j;

	DEBUG( "init_matrix" );

	if( !A || !A->data )
		return( NULL );

	for( i = 0; i < A->height; i++ )
	{
		for( j = 0; j < A->width; j++ )
			A->data[i][j] = val;
	}

	return( A );
}

/****************************************************************/
/*	function	copy_matrix() - copy src matrix to dst			*/
/*																*/
/*	input:		dst	- pointer to destination matrix				*/	
/*				src	- pointer to source matrix					*/	
/*																*/
/*	output:		pointer to destination matrix, or NULL if there */
/*				is an error										*/
/****************************************************************/
MATRIX *copy_matrix( MATRIX *dst, MATRIX *src )
{
	long		i, j;

	DEBUG( "copy_matrix" );

	if( !src || !dst || !src->data || !dst->data || src->width != dst->width ||
												 src->height != dst->height )
		return( NULL );

	for( i = 0; i < src->height; i++ )
	{
		for( j = 0; j < src->width; j++ )
			dst->data[i][j] = src->data[i][j];
	}

	return( dst );
}

/****************************************************************/
/*	function	copy_matrix() - print values in matrix			*/
/*																*/
/*	input:		A	- pointer to matrix to print				*/	
/*																*/
/*	output:		none											*/
/****************************************************************/
void print_matrix( MATRIX *A, char *str )
{
	long		i, j;

	DEBUG( "print_matrix" );

	if( !A )
		return;

	printf( "###############################\n" );
	printf( "%s\n", (str) ? str: "" );
	printf( "width = %ld, height = %ld\n", A->width, A->height );
	printf( "###############################\n" );

	switch( A->type )
	{
		case LONG_TYPE:
				if( !A->ldata )
					return;

				for( i = 0; i < A->height; i++ )
				{
					for( j = 0; j < A->width; j++ )
						printf( "%ld ", A->ldata[i][j] );

					printf( "\n" );
				}
				break;

		case DOUBLE_TYPE:
		default:
				if( !A->data )
					return;

				for( i = 0; i < A->height; i++ )
				{
					for( j = 0; j < A->width; j++ )
						printf( "%5.3f ", A->data[i][j] );

					printf( "\n" );
				}
				break;
	}
}

/****************************************************************/
/*	function	dup_matrix() - allocate new destination matrix, */
/*				and	copy src matrix to it						*/
/*																*/
/*	input:		src	- pointer to source matrix					*/	
/*																*/
/*	output:		pointer to destination matrix, or NULL if there */
/*				is an error										*/
/****************************************************************/
MATRIX *dup_matrix( MATRIX *src )
{
	MATRIX		*dst;

	DEBUG( "dup_matrix" );

	if( !src || !src->data )
		return( NULL );

	if( (dst = alloc_matrix(src->width, src->height, src->type)) == NULL )
		return( NULL );

	return( copy_matrix( dst, src ) );
}

/****************************************************************/
/*	function	matrix_matrix_mult() - multiply 2 matrices		*/
/*				the first matrix is placed on the left, so		*/
/*				its effect is second							*/
/*																*/
/*	input:		pointer to matrix to deallocate					*/
/*																*/
/*	output:		none											*/
/****************************************************************/
MATRIX *matrix_matrix_mult( MATRIX *A, MATRIX *B )
{
	MATRIX		*new;
	long		i, j, k;
	double		tmp;

	DEBUG( "matrix_matrix_mult" );

	if( !A || !B || !A->data || !B->data || A->width != B->height )
		return( NULL );

	if( (new = alloc_matrix(B->width, A->height, A->type)) == NULL )
		return( NULL );

	for( i = 0; i < new->height; i++ )
	{
		for( j = 0; j < new->width; j++ )
		{
			tmp = 0.0;
			for( k = 0; k < A->width; k++ )
				tmp += A->data[i][k] * B->data[k][j];
			new->data[i][j] = tmp;
		}
	}

	return( new );
}

/****************************************************************/
/*	function	scalar_matrix_mult() - multiply a matrix by a 	*/
/*				scalar double value								*/
/*																*/
/*	input:		k	- double scalar to multiply by				*/
/*				A	- pointer to matrix to multiply				*/
/*																*/
/*	output:		pointer to input matrix, or	NULL if input		*/
/*				matrix was NULL to begin with					*/
/****************************************************************/
MATRIX *scalar_matrix_mult( double k, MATRIX *A )
{
	long		i, j;

	DEBUG( "scalar_matrix_mult" );

	if( !A || !A->data )
		return( NULL );

	for( i = 0; i < A->height; i++ )
	{
		for( j = 0; j < A->width; j++ )
			A->data[i][j] *= k;
	}

	return( A );
}

/****************************************************************/
/*	function	transpose_matrix() - allocate new matrix with	*/
/*				same values as input matrix, but with columns	*/
/*				and rows switched								*/
/*																*/
/*	input:		A	- pointer to matrix to transpose			*/
/*																*/
/*	output:		pointer to newly allocated transpose matrix, or	*/
/*				NULL if there is an error						*/
/****************************************************************/
MATRIX *transpose_matrix( MATRIX *A )
{
	MATRIX		*T;
	long		i, j;

	DEBUG( "transpose_matrix" );

	if( !A || !A->data )
		return( NULL );

	if( (T = alloc_matrix(A->height, A->width, A->type)) == NULL )
		return( NULL );

	for( i = 0; i < A->height; i++ )
	{
		for( j = 0; j < A->width; j++ )
			T->data[j][i] = A->data[i][j];
	}

	return( T );
}

/****************************************************************/
/*	function	matrix_matrix_add() - add 2 matrices			*/
/*																*/
/*	input:		A	- pointer to first matrix to add			*/	
/*				B	- pointer to second matrix to add			*/	
/*																*/
/*	output:		pointer to newly allocated sum matrix, or		*/
/*				NULL if there is an error						*/
/****************************************************************/
MATRIX *matrix_matrix_add( MATRIX *A, MATRIX *B )
{
	MATRIX		*S;
	long		i, j;

	DEBUG( "matrix_matrix_add" );

	if( !A || !B || !A->data || !B->data || A->width != B->width ||
													 A->height != B->height )
		return( NULL );

	if( (S = alloc_matrix(A->width, A->height, A->type)) == NULL )
		return( NULL );

	for( i = 0; i < A->height; i++ )
	{
		for( j = 0; j < A->width; j++ )
			S->data[i][j] = A->data[i][j] + B->data[i][j];
	}

	return( S );
}

/****************************************************************/
/*	function	matrix_matrix_sub() - subtract 2 matrices		*/
/*																*/
/*	input:		A	- pointer to first matrix to subtract		*/	
/*				B	- pointer to second matrix to subtract		*/	
/*																*/
/*	output:		pointer to newly allocated difference matrix,	*/
/*				or NULL if there is an error					*/
/****************************************************************/
MATRIX *matrix_matrix_sub( MATRIX *A, MATRIX *B )
{
	MATRIX		*S;
	long		i, j;

	DEBUG( "matrix_matrix_sub" );

	if( !A || !B || !A->data || !B->data || A->width != B->width ||
													 A->height != B->height )
		return( NULL );

	if( (S = alloc_matrix(A->width, A->height, A->type)) == NULL )
		return( NULL );

	for( i = 0; i < A->height; i++ )
	{
		for( j = 0; j < A->width; j++ )
			S->data[i][j] = A->data[i][j] - B->data[i][j];
	}

	return( S );
}

/****************************************************************/
/*	function	normalize_matrix() - normalize the columns of	*/
/*				matrix to have mean = 0 and variance = 1		*/
/*																*/
/*	input:		src	- pointer to matrix to normalize			*/	
/*																*/
/*	output:		pointer to normalized matrix, or NULL if there  */
/*				is an error										*/
/****************************************************************/
MATRIX *normalize_matrix( MATRIX *A )
{
	long		j;

	DEBUG( "normalize_matrix" );

	if( !A || !A->data )
		return( NULL );

	/*
	 * for each column (dimension)
	 */
	for( j = 0; j < A->width; j++ )
		(*column_norm)( A, j );

	return( A );
}

/****************************************************************/
/*	function	matrix_transpose_mult() - multiply matrix by	*/
/*				its own transpose. We could just transpose a	*/
/*				matrix and then call matrix_matrix_mult()		*/
/*				but this avoids having to allocate a separate	*/
/*				matrix for the transpose, and since we'll be	*/
/*				dealing with BIG matrices - this is a good 		*/
/*				thing											*/
/*																*/
/*	input:		pointer to matrix to "square"					*/
/*																*/
/*	output:		pointer to newly allocated multiplied matrix, 	*/
/*				or NULL of there is an error					*/
/****************************************************************/
MATRIX *matrix_transpose_mult( MATRIX *A )
{
	MATRIX		*T;
	long		i, j, k;
	double		tmp;

	DEBUG( "matrix_transpose_mult" );

	if( !A || !A->data )
		return( NULL );

	if( (T = alloc_matrix(A->width, A->width, A->type)) == NULL )
		return( NULL );

	for( i = 0; i < T->height; i++ )
	{
		for( j = 0; j < T->width; j++ )
		{
			tmp = 0.0;
			for( k = 0; k < A->width; k++ )
				tmp += A->data[k][i] * A->data[k][j];
			T->data[i][j] = tmp;
		}
	}

	return( T );
}

/****************************************************************/
/*	function	correlation_matrix() - calculate correlation	*/
/*				matrix of an input matrix						*/
/*																*/
/*	input:		src	- pointer to matrix for which to compute	*/
/*				correlation matrix								*/	
/*																*/
/*	output:		pointer to newly allocated correlation matrix, 	*/
/*				or NULL if there is an error					*/
/****************************************************************/
MATRIX *correlation_matrix( MATRIX *A )
{
	MATRIX		*R;

	DEBUG( "correlation_matrix" );

	if( !A || !A->data )
		return( NULL );

	if( (R = matrix_transpose_mult(A)) == NULL )
		return( NULL );

	return( scalar_matrix_mult( 1.0/A->height, R) );
}

/****************************************************************/
/*	function	distance_matrix() - allocate new matrix with	*/
/*				cell i,j containing distance between ith, jth	*/
/*				data points										*/
/*																*/
/*	input:		A	- pointer to matrix to calc distances for	*/
/*																*/
/*	output:		pointer to newly allocated distance matrix, or	*/
/*				NULL if there is an error						*/
/****************************************************************/
MATRIX *distance_matrix( MATRIX *A, int mask[] )
{
	MATRIX		*D;
	long		i, j;

	DEBUG( "distance_matrix" );

	if( !A || !A->data )
		return( NULL );

	if( (D = alloc_matrix(A->height, A->height, A->type)) == NULL )
		return( NULL );

	for( i = 0; i < D->height; i++ )
	{
		for( j = 0; j < i; j++ )
			D->data[i][j] = (*distance)(A->data[j],A->data[i],mask,A->width);
	}

	return( D );
}

/****************************************************************/
/*  function:   column_mean() - calculate mean of values in a   */
/*              given column (dimension) of a data matrix		*/
/*              Uses formula 14.1.1, Numerical Recipies in C:  	*/
/*                               --                             */
/*                  mean =   1/n \  x                           */
/*                               /                              */
/*                               --                             */
/*                              x in X                          */
/*                                                              */
/*  input:      A       - pointer to matrix						*/
/*              dim     - integer index - dimension of interest */
/*                                                              */
/*  output:     value of mean for given dimension               */
/****************************************************************/
double column_mean( MATRIX *A, long j )
{
	long	i;
	double	sum = 0.0;

	DEBUG( "column_mean" );

	if( !A || !A->data || j < 0 || j >= A->width )
		return( 0.0 );

	for( i = 0; i < A->height; i++ )
		sum += A->data[i][j];

	return( sum/A->height );
}

/****************************************************************/
/*  function:   column_variance() - calculate variance of 		*/
/*              values in a given column (dimension) of a data 	*/
/*              matrix.  Uses formula 14.1.2, Numerical 		*/
/*				Recipies in C:  								*/
/*                                       --                     */
/*                  variance =   1/(n-1) \  (x - X_mean)^2      */
/*                                       /                      */
/*                                       --                     */
/*                                      x in X                  */
/*                                                              */
/*  input:      A       - pointer to matrix						*/
/*              dim     - integer index - dimension of interest */
/*                                                              */
/*  output:     value of variance for given dimension           */
/****************************************************************/
double column_variance( MATRIX *A, long j )
{
	long	i;
	double	val, mean;
	double	sum = 0.0;

	DEBUG( "column_variance" );

	if( !A || !A->data || j < 0 || j >= A->width )
		return( 0.0 );

	/*
     * calculate mean for this dimension
     */
    mean = column_mean( A, j );

	for( i = 0; i < A->height; i++ )
	{
		val = A->data[i][j] - mean;
		sum += SQR(val);
	}

	return( sum/(A->height-1) );
}

/****************************************************************/
/*  function:   column_median() - calculate median of values in	*/
/*              a given column (dimension) of a data matrix.  	*/
/*																*/
/*  input:      A       - pointer to matrix						*/
/*              dim     - integer index - dimension of interest */
/*																*/
/*	output:		value of median for given dimension				*/
/****************************************************************/
double column_median( MATRIX *A, long j )
{
	int		*tmp_mask;
	double	*tmp_vect;
	double 	median;
	long	i;
	
	DEBUG( "column_median" );

	/*
	 * check that dimension is in valid range
	 */
	if( !A || !A->data || j < 0 || j >= A->width )
		return( 0.0 );

	if( (tmp_vect = alloc_nvect(A->height)) == NULL )
		return( 0.0 );
	
	if( (tmp_mask = (int *)mmalloc(A->height)) == NULL )
		return( 0.0 );
	
	/*
	 * collect values for this dimension in vector
	 */	
	for( i = 0; i < A->height; i++ )
	{
		tmp_vect[i] = A->data[i][j];
		tmp_mask[i] = 1;
	}
	
	/*
	 * find median for this dimension
	 */
	median = median_nvect( tmp_vect, tmp_mask, A->height );
	
	mfree( tmp_vect );
	mfree( tmp_mask );
	
	return( median );
}

/****************************************************************/
/*  function:   column_min() - calculate min of values in		*/
/*              a given column (dimension) of a data matrix.  	*/
/*																*/
/*  input:      A       - pointer to matrix						*/
/*              dim     - integer index - dimension of interest */
/*																*/
/*	output:		value of min for given dimension				*/
/****************************************************************/
double column_min( MATRIX *A, long j )
{
	double 	min = DBL_MAX;
	long	i;
	
	DEBUG( "column_min" );

	/*
	 * check that dimension is in valid range
	 */
	if( !A || !A->data || j < 0 || j >= A->width )
		return( 0.0 );

	/*
	 * collect values for this dimension in vector
	 */	
	for( i = 0; i < A->height; i++ )
		min = MIN(min, A->data[i][j]);
	
	return( min );
}

/****************************************************************/
/*  function:   column_max() - calculate max of values in		*/
/*              a given column (dimension) of a data matrix.  	*/
/*																*/
/*  input:      A     - pointer to matrix						*/
/*              j     - integer index - dimension of interest 	*/
/*																*/
/*	output:		value of max for given dimension				*/
/****************************************************************/
double column_max(  MATRIX *A, long j )
{
	double 	max = DBL_MIN;
	long	i;
	
	DEBUG( "column_max" );

	/*
	 * check that dimension is in valid range
	 */
	if( !A || !A->data || j < 0 || j >= A->width )
		return( 0.0 );

	/*
	 * collect values for this dimension in vector
	 */	
	for( i = 0; i < A->height; i++ )
		max = MAX(max, A->data[i][j]);
	
	return( max );
}

/****************************************************************/
/*	function:	column_noop() - do nothing :-) This allows us 	*/
/*				to let user apply no normalization to data		*/
/*																*/
/*	input:		A 		- pointer to matrix to normalize		*/
/*				j		- integer index - dimension of interest */	
/*																*/ 
/*	output:		none											*/	
/****************************************************************/
void column_noop( MATRIX *A, long j )
{
	DEBUG( "column_noop" );
}

/****************************************************************/
/*	function:	column_scale() - scale values in givendimension	*/
/*				(i.e. column of data matrix) so	that they are 	*/
/*				0.0 - 1.0						 				*/
/*																*/
/*	input:		A 		- pointer to matrix to normalize		*/
/*				j		- integer index - dimension of interest */	
/*																*/
/*	output:		none											*/	
/****************************************************************/
void column_scale( MATRIX *A, long j )
{
	double	min, max;
	double	range;
	double	inv_range;
	long	i;
	
	DEBUG( "column_scale" );

	/*
	 * check that dimension is in valid range
	 */
	if( !A || !A->data || j < 0 || j >= A->width )
		return;
		
	/*
	 * find minimum and maximum of values in column
	 */
	min = column_min( A, j );
	max = column_max( A, j );
	
	range = max - min;
	
	if( FLT_ZERO(range) )
		return;

	inv_range = 1.0/range;
	
	/*
	 * scale all values in column to be 0.0 - 1.0
	 */	
	for( i = 0; i < A->height; i++ )
		A->data[i][j] = (A->data[i][j] - min)*inv_range;

}

/****************************************************************/
/*	function:	column_submean() - subtract "feature" mean from */
/*				all values in a given dimension	(i.e. column)	*/
/*				of data matrix									*/
/*																*/
/*	input:		A 		- pointer to matrix to normalize		*/
/*				j		- integer index - dimension of interest */	
/*																*/
/*	output:		none											*/	
/****************************************************************/
void column_submean( MATRIX *A, long j )
{
	double	mean;
	long	i;
	
	DEBUG( "column_submean" );

	/*
	 * check that dimension is in valid range
	 */
	if( !A || !A->data || j < 0 || j >= A->width )
		return;
		
	mean = column_mean( A, j );
	
	/*
	 * subtract mean from all values in given dimension
	 */	
	for( i = 0; i < A->height; i++ )
		A->data[i][j] -= mean;
}

/****************************************************************/
/*	function:	column_unitvar() - scale values in given dim	*/
/*				(i.e. column of data matrix) so that they have 	*/
/*				mean = 0, and variance = 1						*/
/*																*/
/*	input:		A 		- pointer to matrix to normalize		*/
/*				j		- integer index - dimension of interest */	
/*																*/
/*	output:		none											*/	
/****************************************************************/
void column_unitvar( MATRIX *A, long j )
{
	double	mean;
	double	variance;
	double	st_dev;
	double	factor;
	long	i;
	
	DEBUG( "column_unitvar" );

	/*
	 * check that dimension is in valid range
	 */
	if( !A || !A->data || j < 0 || j >= A->width )
		return;
		
	mean = column_mean( A, j );
	
	/*
	 * subtract mean from all values in given dimension
	 */	
	for( i = 0; i < A->height; i++ )
		A->data[i][j] -= mean;
	
	variance = column_variance( A, j );
	st_dev = sqrt( variance );

	if( FLT_ZERO(st_dev) )
		return;

	factor = 1.0/st_dev;
	
	/*
	 * scale all values in dimension by standard deviation
	 */	
	for( i = 0; i < A->height; i++ )
		A->data[i][j] *= factor;

}

/****************************************************************/
/*	function:	calc_dim_correlation() - calculates the			*/
/*				correlation between two columns (dimensions)	*/
/*				in a data matrix. Uses formula for correlation, */
/*				"Pearson's r" from	Numerical Recipies (14.5.1)	*/
/*				(p. 636):										*/
/*																*/
/*				         --					--					*/	
/*						 \ (x - X_mean) * 	\ (y - Y_mean)		*/
/*				         /					/					*/
/*				         --					-- 					*/
/*						x in X				y in Y				*/
/*				r = ---------------------------------------		*/
/*				   ------------------      -----------------	*/
/*			       | --					   | --					*/	
/*			    -  | \ (x - X_mean)^2 * -  | \ (y - Y_mean)^2	*/
/*			     \ | /					 \ | /					*/
/*			      \| --					  \| -- 				*/
/*					 x in X					 y in Y				*/
/*																*/
/*	input:		A 		- pointer to matrix to operate on		*/
/*				dim1	- integer index - dimension of interest */	
/*				dim2	- integer index - dimension of interest */	
/*																*/
/*	output:		double correlation between distances for db		*/	
/****************************************************************/
double calc_dim_correlation( MATRIX *A, long dim1, long dim2 )
{
	double	dim1_mean, dim2_mean;
	double	dim1_delta, dim2_delta;
	double	dim1_sum = 0;
	double	dim2_sum = 0;
	double	numer = 0;
	double	denom = 0;
	long	i;
	
	DEBUG( "calc_dim_correlation" );

	if( !A || !A->data || dim1 < 0 || dim1 >= A->width || 
									dim2 < 0 || dim2 >= A->width )
		return( 0.0 );
		
	/*
	 * calculate mean of display-dimensional distances
	 * mean of n-dimensional distances have already been
	 * computed in calc_all_distances
	 */
	dim1_mean = column_mean( A, dim1 );
	dim2_mean = column_mean( A, dim2 );
	
	/*
	 * sum values for each dimension
	 */
	for( i = 0; i < A->height; i++ )
	{
		dim1_delta = (A->data[i][dim1] - dim1_mean);
		dim2_delta = (A->data[i][dim2] - dim2_mean);
		
		numer += (dim1_delta * dim2_delta);
		
		dim1_sum += SQR(dim1_delta);
		dim2_sum += SQR(dim2_delta);
	}
	
	denom = sqrt(dim1_sum) * sqrt(dim2_sum);
	
	return( (FLT_ZERO(denom)) ? 0.0 : numer/denom );
}

MATRIX *matrix_square( MATRIX *A )
{
	MATRIX	*S;
	long	i, j;

	DEBUG( "matrix_square" );

	if( !A || !A->data )
		return( NULL );

	if( (S = alloc_matrix(A->width, A->height, A->type)) == NULL )
		return( NULL );

	for( i = 0; i < A->height; i++ )
	{
		for( j = 0; j < A->width; j++ )
			S->data[i][j] = SQR(A->data[i][j]);
	}

	return( S );
}

double matrix_summation( MATRIX *A )
{
	double	sum = 0.0;
	long	i, j;

	DEBUG( "matrix_summation" );

	if( !A || !A->data )
		return( 0.0 );

	for( i = 0; i < A->height; i++ )
	{
		for( j = 0; j < A->width; j++ )
			sum += A->data[i][j];
	}

	return( sum );
}

double matrix_mean( MATRIX *A )
{
	DEBUG( "matrix_mean" );

	if( A->height == 0 || A->width == 0 )
		return( 0.0 );

	return( matrix_summation(A)/(A->height*A->width) );
}

/****************************************************************/
/*  function:   dist_correlation() - calculates the        		*/
/*              correlation between the display-dimensional     */
/*              and the n-dimensional distances for the entire  */
/*              database. Uses formula for correlation,         */
/*              "Pearson's r" from  Numerical Recipies (14.5.1) */
/*              (p. 636):                                       */
/*                                                              */
/*                       --                 --                  */
/*                       \ (x - X_mean) *   \ (y - Y_mean)      */
/*                       /                  /                   */
/*                       --                 --                  */
/*                      x in X              y in Y              */
/*              r = ---------------------------------------     */
/*                          X_std_dev * Y_std_dev               */
/*                                                              */
/*  input:      db      - pointer to database                   */
/*                                                              */
/*  output:     double correlation between distances for db     */
/****************************************************************/
double dist_correlation( MATRIX *A, MATRIX *P, int mask[] )
{
	long	i, j;
	double	ndist, sdist;
    double  ndelta, sdelta;
    double  denom_nsum = 0;
    double  denom_ssum = 0;
    double  numer = 0;
    double  denom;
    double  sdist_mean;
	double 	ndist_mean;

	DEBUG( "dist_correlation" );

	if( !A || !P || !A->data || !P->data || A->height != P->height )
		return( 0.0 );

	/*
     * calculate mean of display-dimensional distances
     */
	ndist_mean = calc_mean_distance( A, mask );
    sdist_mean = calc_mean_distance( P, display_mask );

	for( i = 0; i < A->height; i++ )
	{
		for( j = 0; j < i; j++ )
		{
			ndist = (*distance)( A->data[j], A->data[i], mask, A->width );
			sdist = (*distance)( P->data[j], P->data[i], display_mask,P->width);

            ndelta = (ndist - ndist_mean);
            sdelta = (sdist - sdist_mean);

            numer += (ndelta * sdelta);

            denom_nsum += SQR(ndelta);
            denom_ssum += SQR(sdelta);
        }
    }

	denom = sqrt(denom_nsum) * sqrt(denom_ssum);
	
	return( (FLT_ZERO(denom)) ? 0.0 : numer/denom );
}

/****************************************************************/
/*  function    calc_max_distances() - calculate the "farthest" */
/*              two data points and passes their indexes back   */
/*              to the calling function                         */
/*                                                              */
/*  input:      A      - pointer to matrix						*/
/*              index1 - index of datum pointer to receive   	*/
/*                       location of 1st of "farthest" datums  	*/
/*              index2 - index of datum pointer to receive   	*/
/*                       location of 2nd of "farthest" datums  	*/
/*																*/
/*  output:     maximum distance between data points            */
/****************************************************************/
double calc_max_distance( MATRIX *A, int mask[], long *index1, long *index2 )
{
	long        i, j;
    double      ndist, max_ndist;

	DEBUG( "calc_max_distance" );

	max_ndist = (*distance)( A->data[0], A->data[1], mask, A->width );
	*index1 = 0;
	*index2 = 1;

	for( i = 0; i < A->height; i++ )
	{
		for( j = 0; j < i; j++ )
		{
			ndist = (*distance)( A->data[i], A->data[j], mask, A->width );

            if( ndist > max_ndist )
            {
                max_ndist = ndist;
                *index1 = j;
                *index2 = i;
            }  
		}
	}

	return( max_ndist );
}

/****************************************************************/
/*  function    calc_min_distances() - calculate the "nearest"  */
/*              two data points and passes their indexes back   */
/*              to the calling function                         */
/*                                                              */
/*  input:      A      - pointer to matrix						*/
/*              index1 - index of datum pointer to receive   	*/
/*                       location of 1st of "nearest" datums   	*/
/*              index2 - index of datum pointer to receive   	*/
/*                       location of 2nd of "nearest" datums   	*/
/*																*/
/*  output:     minimum distance between data points            */
/****************************************************************/
double calc_min_distance( MATRIX *A, int mask[], long *index1, long *index2 )
{
	long        i, j;
    double      ndist, min_ndist;

	DEBUG( "calc_min_distance" );

	min_ndist = (*distance)( A->data[0], A->data[1], mask, A->width );
	*index1 = 0;
	*index2 = 1;

	for( i = 0; i < A->height; i++ )
	{
		for( j = 0; j < i; j++ )
		{
			ndist = (*distance)( A->data[i], A->data[j], mask, A->width );

            if( ndist < min_ndist )
            {
                min_ndist = ndist;
                *index1 = j;
                *index2 = i;
            }  
		}
	}

	return( min_ndist );
}

/****************************************************************/
/*  function    calc_mean_distance() - calculate the mean		*/
/*				distance between data points in a matrix		*/
/*                                                              */
/*  input:      A       - pointer to matrix						*/
/*				mask	- array of integer enable/disable flags	*/
/*																*/
/*  output:     minimum distance between data points            */
/****************************************************************/
double calc_mean_distance( MATRIX *A, int mask[] )
{
	long        i, j, count = 0;
    double      dist_sum = 0.0;

	DEBUG( "calc_mean_distance" );

	for( i = 0; i < A->height; i++ )
	{
		for( j = 0; j < i; j++ )
		{
			dist_sum += (*distance)( A->data[i], A->data[j], mask, A->width );
			count++;
		}
	}

	return( dist_sum/(double)count );
}

double kruskal_stress( MATRIX *A, MATRIX *P, int mask[] )
{
    long    i, j;
	double	ndist, sdist;
	double	ndist_sum = 0.0;
	double	error_sum = 0.0;
	double	err;

	if( !A || !P || !A->data || !P->data || A->height != P->height )
		return( 0.0 );

	DEBUG( "kruskal_stress" );

	for( i = 0; i < A->height; i++ )
    {
        for( j = 0; j < i; j++ )
		{
			ndist = (*distance)( A->data[j], A->data[i], mask, A->width );
			sdist = (*distance)( P->data[j], P->data[i], display_mask,P->width);

			err = ndist - sdist;
			error_sum += SQR(err);
			ndist_sum += SQR(ndist);
		}
	}

	return( FLT_ZERO(ndist_sum)? 0.0 : sqrt(error_sum/ndist_sum) );
}

double sammon_stress( MATRIX *A, MATRIX *P, int mask[] )
{
    long    i, j;
	double	ndist, sdist;
	double	ndist_sum = 0.0;
	double	error_sum = 0.0;
	double	err;

	DEBUG( "sannoB_stress" );

	if( !A || !P || !A->data || !P->data || A->height != P->height )
		return( 0.0 );

	for( i = 0; i < A->height; i++ )
    {
        for( j = 0; j < i; j++ )
		{
			ndist = (*distance)( A->data[j], A->data[i], mask, A->width );
			sdist = (*distance)( P->data[j], P->data[i], display_mask,P->width);

			err = ndist - sdist;
			error_sum += SQR(err)/ndist;
			ndist_sum += ndist;
		}
	}

	return( FLT_ZERO(ndist_sum) ? 0.0 : error_sum/ndist_sum );
}

double lingoes_stress( MATRIX *A, MATRIX *P, int mask[] )
{
    long    i, j;
	double	ndist, sdist;
	double	error_sum = 0.0;
	double	err;

	DEBUG( "lingoes_stress" );

	if( !A || !P || !A->data || !P->data || A->height != P->height )
		return( 0.0 );

	for( i = 0; i < A->height; i++ )
    {
        for( j = 0; j < i; j++ )
		{
			ndist = (*distance)( A->data[j], A->data[i], mask, A->width );
			sdist = (*distance)( P->data[j], P->data[i], display_mask,P->width);

			err = sdist - ndist;
			error_sum += ABS(err)/sdist;
		}
	}

	return( error_sum );
}

double J_ee_stress( MATRIX *A, MATRIX *P, int mask[] )
{
    long    i, j;
	double	ndist, sdist;
	double	ndist_sum = 0.0;
	double	error_sum = 0.0;
	double	err;

	DEBUG( "J_ee_stress" );

	if( !A || !P || !A->data || !P->data || A->height != P->height )
		return( 0.0 );

	for( i = 0; i < A->height; i++ )
    {
        for( j = 0; j < i; j++ )
		{
			ndist = (*distance)( A->data[j], A->data[i], mask, A->width );
			sdist = (*distance)( P->data[j], P->data[i], display_mask,P->width);

			err = sdist - ndist;
			error_sum += SQR(err);
			ndist_sum += SQR(ndist);
		}
	}

	return( FLT_ZERO(ndist_sum) ? 0.0 : error_sum/ndist_sum );
}

double J_ff_stress( MATRIX *A, MATRIX *P, int mask[] )
{
    long    i, j;
	double	ndist, sdist;
	double	error_sum = 0.0;
	double	err;

	DEBUG( "J_ff_stress" );

	if( !A || !P || !A->data || !P->data || A->height != P->height )
		return( 0.0 );

	for( i = 0; i < A->height; i++ )
    {
        for( j = 0; j < i; j++ )
		{
			ndist = (*distance)( A->data[j], A->data[i], mask, A->width );
			sdist = (*distance)( P->data[j], P->data[i], display_mask,P->width);

			err = (sdist - ndist)/ndist;
			error_sum += SQR(err);
		}
	}

	return( error_sum );
}

double J_ef_stress( MATRIX *A, MATRIX *P, int mask[] )
{
    long    i, j;
	double	ndist, sdist;
	double	ndist_sum = 0.0;
	double	error_sum = 0.0;
	double	err;

	DEBUG( "J_ef_stress" );

	if( !A || !P || !A->data || !P->data || A->height != P->height )
		return( 0.0 );

	for( i = 0; i < A->height; i++ )
    {
        for( j = 0; j < i; j++ )
		{
			ndist = (*distance)( A->data[j], A->data[i], mask, A->width );
			sdist = (*distance)( P->data[j], P->data[i], display_mask,P->width);

			err = sdist - ndist;
			error_sum += SQR(err)/ndist;
			ndist_sum += ndist;
		}
	}

	return( FLT_ZERO(ndist_sum) ? 0.0 : error_sum/ndist_sum );
}

double max_distance_from_i( MATRIX *A, int mask[], long i, long *j_index )
{
    long        j;
    double      ndist;
    double      max_ndist = DBL_MIN;

    DEBUG( "calc_max_distance" );

    for( j = 0; j < i; j++ )
    {
        ndist = (*distance)( A->data[i], A->data[j], mask, A->width );

        if( ndist > max_ndist )
        {
            max_ndist = ndist;
            *j_index = j;
        }
    }

    return( max_ndist );
}

double min_distance_from_i( MATRIX *A, int mask[], long i, long *j_index )
{
    long        j;
    double      ndist;
    double      min_ndist = DBL_MAX;

    DEBUG( "calc_min_distance" );

    for( j = 0; j < i; j++ )
    {
        ndist = (*distance)( A->data[i], A->data[j], mask, A->width );

        if( ndist < min_ndist )
        {
            min_ndist = ndist;
            *j_index = j;
        }
    }

    return( min_ndist );
}
