/************************************************************************/
/*  headers																*/
/************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "general.h"
#include "util.h"
#include "viz.h"
#include "nmatrix.h"
#include "animate.h"

/************************************************************************/
/*  defines																*/
/************************************************************************/
#define reset_interpolate(db)	(db)->dat.num_P = 0

/************************************************************************/
/*  typedefs															*/
/************************************************************************/

/************************************************************************/
/*  function prototypes													*/
/************************************************************************/
void animate_db( void *data );
void interpolate_CatmullRom( DB *db );
void cycle_display_matrices( DB *db );

/************************************************************************/
/*  external functions 													*/
/************************************************************************/

/************************************************************************/
/*  global variables													*/
/************************************************************************/
char buf[1024];


/****************************************************************/
/*	function:	animate_db() - repeadedly compute a new screen	*/
/*				position for each datum, based on its "gravity"	*/
/*				to other datums, and where those other datums	*/
/*				are on the screen. Here we do not recompute the	*/
/*				bonds for each pair of datums, but rely on		*/
/*				the top list of bonds which we precomputed		*/
/*																*/
/*	input:		db 		- pointer to database					*/
/*																*/
/*	output:		none											*/	
/****************************************************************/
void animate_db( void *data )
{
	DB		*db = (DB *)data;
//	DATUM	*datum;
	
	DEBUG( "animate_db" );

	if( !db || !db->dat.data )
		return;

	/*
	 * clear once to ensure we start off new animation
	 * draw_db may not clear if TRACE is enabled
	 */
//	reset_correlate( db, FALSE );
//	reset_stress( db, FALSE );

	reset_interpolate( db );

	for( db->animate.cur_iteration = 0; 
		 db->animate.cur_iteration < db->animate.max_iteration; 
		 db->animate.cur_iteration++ )
	{
//		update_counter( db->animate.cur_iteration );

		if( db->animate.interpolate  )
			cycle_display_matrices( db );

		/*
		 * for each datum, get its bonds to all the
		 * points on its top list, and use this to calculate
		 * a delta for the datum to move in screen coords
		 */
		calc_all_positions( db );

#ifdef PDEBUG
		print_matrix( db->dat.sD, "sD" );
		print_matrix( db->dat.P, "P" );
		print_matrix( db->dat.O, "O" );
#endif

		
		if( db->animate.cur_iteration == 0 )
		{
			db->stats.start_max_bond_stress = db->stats.max_bond_stress;
			db->stats.start_max_datum_stress = db->stats.max_datum_stress;
		}

//		if( db->animate.interpolate  )
//			interpolate_CatmullRom( db );
//		else
//			draw_db( db );
	}
}

void cycle_display_matrices( DB *db )
{
	MATRIX	*tmp;

	DEBUG( "cycle_display_matrices" );

	tmp = db->dat.POOO;
	db->dat.POOO = db->dat.POO;
	db->dat.POO = db->dat.PO;
	db->dat.PO = tmp;

	if( db->dat.num_P < 4 )
		db->dat.num_P++;
}

void interpolate_CatmullRom( DB *db )
{
	VECT	uvect, u_M, pt;
	double	u, sqr_u, du;
	long	i;
	MATRIX	*tmp;
	MAT4X4	P;
	MAT4X4	M = { { -1,  2, -1,  0 },
				  {  3, -5,  0,  2 },
				  { -3,  4,  1,  0 },
				  {  1, -1,  0,  0 },
				};
 
	DEBUG( "interpolate_CatmullRom" );

	if( db->dat.num_P < 4 )
		return;

	/*
	 * setup Catmull-Rom Basis matrix
	 */
	scalar_mat_mult( 0.5, M );
	zero_mat( P );

	du = 1.0/(db->animate.interp_steps-1);

	/*
	 * interpolate splines for various increments of u (0.0 - 1.0)
	 */
	for( u = 0; u <= 1.0; u += du )
	{
		/*
		 * set up power vector
		 */
		sqr_u = SQR(u);
		set_vect( uvect, u*sqr_u, sqr_u, u, 1 );

		/*
		 * premultiply Basis matrix and power vector
		 */
		pt_mat_mult( uvect, M, u_M );

		/*
		 * interpolate all points in configuration by
		 * current value of u
		 */
		for( i = 0; i < db->dat.P->height; i++ )
		{
			/*
			 * set up matrix of control points (oldest first)
			 */
			zero_mat( P );
			load_mat_cols( P, db->dat.POOO->data[i], db->dat.POO->data[i],
						      db->dat.PO->data[i], db->dat.P->data[i],
							  db->dat.P->width );

			/*
			 * multiply Catmull-Rom matrix by input point
			 * and store interpolated point in 'IP' matrix
			 */
			pt_mat_mult( u_M, P, pt );
			copy_nvect( db->dat.IP->data[i], pt, db->dat.P->width );
		}

		/*
		 * temporarily swap IP for P so draw_db 
		 * will draw interpolated points. Then
		 * swap back - so IP will get reused
	 	 */
		tmp = db->dat.P;
		db->dat.P = db->dat.IP;

//		draw_db( db );

		db->dat.P = tmp;
	}
}
