/****************************************************************/
/* headers                                                      */
/****************************************************************/
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <limits.h>
#include "viz.h"
#include "util.h"
#include "nvect.h"
#include "nmatrix.h"
#include "bonds.h"

/****************************************************************/
/* defines                                                      */
/****************************************************************/
#define ADD_BOND_NEAR( i, j, ndist, off_dist, offj ) 						  \
offj = -1, (*add_bond)( nD->data[(i)], O->ldata[(i)], 0, t, (j), (ndist), 	  \
			 &db->dat.data[(i)].num_bonds, &(off_dist), &(offj), is_nearer)   \


#define ADD_BOND_FAR( i, j, ndist, off_dist, offj ) 						  \
offj = -1, (*add_bond)( nD->data[(i)], O->ldata[(i)], t, limit, (j), (ndist), \
			 &db->dat.data[(i)].num_bonds, &(off_dist), &(offj), is_further)  \

#define BOND_DROPPED( off )		((off) != -1)

/****************************************************************/
/* typedefs                                                     */
/****************************************************************/
typedef int (*TST_FUNC)( double, double );
		
/****************************************************************/
/* function prototypes                                          */
/****************************************************************/
/*
 * interface functions
 */
void calc_all_distances( DB *db );
void calc_all_positions( DB *db );
void set_stress_func( DB *db );

/*
 * private functions
 */
static void calc_all_ndist( DB *db, MATRIX *A, MATRIX *nD, MATRIX *O, 
							long limit, int mask[], int mode, MDS_STATS *stats);
static void calc_all_ndist_random( DB *db, MATRIX *A, MATRIX *nD, MATRIX *O, 
						    long limit, int mask[], MDS_STATS *stats );
void calc_all_sdist( DB *db, MATRIX *P, MATRIX *O, MATRIX *sD );
void calc_all_stress( DB *db, MATRIX *nD, MATRIX *sD, MATRIX *S );
int add_bond_sorted( double dists[], long others[], long low,long high, 
						    long j, double ndist, long *num_bonds, 
							double *off_dist, long *offj, TST_FUNC tst );
static void calc_new_positions( DB *db, MATRIX *OP, MATRIX *NP, MATRIX *S, 
							MATRIX *O, double step_fact, double noise_fact);
int is_nearer( double new_dist, double old_dist );
int is_further( double new_dist, double old_dist );

/*
 * stress calculating functions
 */
static double calc_stress_lingoes( double ndist, double sdist, 
								double ndist_sum, double sdist_sum );
static double calc_stress_sammon( double ndist, double sdist,
								double ndist_sum, double sdist_sum );
static double calc_stress_J_ee( double ndist, double sdist, 
								double ndist_sum, double sdist_sum );
static double calc_stress_J_ff( double ndist, double sdist, 
								double ndist_sum, double sdist_sum );
static double calc_stress_J_ef( double ndist, double sdist, 
								double ndist_sum, double sdist_sum );

/****************************************************************/
/* external functions											*/
/****************************************************************/
extern void set_global_stress( int mode );

/****************************************************************/
/* global variables                                          	*/
/****************************************************************/
double (*calc_stress)( double, double, double, double ) = calc_stress_lingoes;
int (*add_bond)( double [], long [], long, long, long, double, 
				 long *, double *, long *, TST_FUNC ) = add_bond_sorted;

/****************************************************************/
/*	function	calc_all_distances() - for every data point 	*/
/*				in data base, calculate its distance to every 	*/
/*				other point. Also calculate min, max, mean, and */
/*				standard deviation of n-dimensional distances,	*/
/*				and save these statistics in the database		*/
/*				structure										*/
/*																*/
/*	input:		db 		- pointer to database					*/
/*	output:		none											*/
/****************************************************************/
void calc_all_distances( DB *db )
{
	DEBUG( "calc_all_distances" );

	/*
	 * setup appropriate function pointers
	 */
	set_distance_func( db->mds.dist_func );

	/*
	 * calculate all the pairwise n-dimensional
	 * distances between data points
	 */
	switch( db->mds.select_mode )
	{
		case TOP_TEN:
				calc_all_ndist( db, db->dat.A, db->dat.nD, 
								db->dat.O, db->mds.num_bonds,
								db->dat.dim_mask, db->mds.select_mode, 
								&db->stats );
				break;
		
		case RANDOM_ONCE:
		default:
				calc_all_ndist_random(  db, db->dat.A, db->dat.nD,
										db->dat.O, db->mds.num_bonds,
										db->dat.dim_mask, &db->stats );
				break;
	}
}

/****************************************************************/
/*	function	calc_all_positions() - for every data point		*/
/*				in data base, calculate its new position in 	*/
/*				display coordinates								*/
/*																*/
/*	input:		db 		- pointer to database					*/
/*	output:		none											*/
/****************************************************************/
void calc_all_positions( DB *db )
{
	DEBUG( "calc_all_positions" );

	/*
	 * setup STRESS function pointers
	 */
	set_stress_func( db );

	/*
	 * calculate all distances between points in screen space
	 * and then calculate stress between two representations
	 */
	calc_all_sdist( db, db->dat.P, db->dat.O, db->dat.sD );
	calc_all_stress( db, db->dat.nD, db->dat.sD, db->dat.S );
	
	/*
	 * save old point positions
	 */
	copy_matrix( db->dat.PO, db->dat.P );

	/*
	 * for each data point update its screen
	 * position based on its stress with other
	 * points that it's tracking (bonded with)
	 */
	calc_new_positions( db, db->dat.PO, db->dat.P, db->dat.S, db->dat.O,
							db->mds.step_factor, db->mds.noise_factor );
}

/****************************************************************/
/*	function	set_stress_func() - select flavor of STRESS to	*/
/*				calculate for distance relations				*/
/*																*/
/*	input:		db 		- pointer to database					*/
/*																*/
/*	output:		none											*/
/****************************************************************/
void set_stress_func( DB *db )
{
	DEBUG( "set_stress_func" );

	switch( db->mds.stress_function )
	{
		case J_EE_STRESS:
			calc_stress = calc_stress_J_ee;
			break;
		case J_FF_STRESS:
			calc_stress = calc_stress_J_ff;
			break;
		case J_EF_STRESS:
			calc_stress = calc_stress_J_ef;
			break;
		case SAMMON_STRESS:
			calc_stress = calc_stress_sammon;
			break;
		case KRUSKAL_STRESS1:
		case KRUSKAL_STRESS2:
		case LINGOES_STRESS:
		default:
			calc_stress = calc_stress_lingoes;
			break;
	}

	/*
	 * also set the function pointer used to
	 * calculate the global stress - i.e. the
	 * stress for all the n(n-1)/2 distance
	 * relationships - not just the n*p subset
	 * distances
	 */
	set_global_stress( db->mds.stress_function );
}

/****************************************************************/
/*	function	calc_all_ndist() - for every point in data base */
/*				calculate its distance to every other point. 	*/
/*				Then for each point save the top NUM_ATTRACTION */
/*				points which exert the greatest effect on it.  	*/
/*				Also calculate min, max, mean, standard dev		*/
/*				of n-dimensional distances,	and save these 		*/
/*				statistics in the database structure			*/
/*																*/
/*	input:		db 		- pointer to database					*/
/*	output:		none											*/
/****************************************************************/
static void calc_all_ndist( DB *db, MATRIX *A, MATRIX *nD, MATRIX *O, 
							long limit, int mask[], int mode, MDS_STATS *stats )
{
	long		offj, offj2;
	long		i, j, t, n = 0;
	double		off_dist, off_dist2;
	double		min = DBL_MAX;
	double		max = DBL_MIN;
	double		sum = 0.0;
	double		mean = 0.0;
	double		var = 0.0;
	double		ndist; 
	
	DEBUG( "calc_all_ndist" );

	stats->num_dropped = 0;
	t = limit/2;

	/*
	 * calculate all n(n-1)/2 distances between data points
	 * also calculate mean, min, max n-dimensional distances
	 */	
	for( i = 0; i < A->height; i++ )
	{
		db->dat.data[i].num_bonds = 0;

		for( j = 0; j < i; j++ )
		{
			ndist = (*distance)( A->data[i], A->data[j], mask, A->width);

			sum += ndist;
			min = MIN( ndist, min );
			max = MAX( ndist, max );
			
			if( ADD_BOND_NEAR( i, j, ndist, off_dist, offj ) )
			{
				/*
				 * first try to add distance from i -> j to i's list
				 * see if it is a distance which belongs among the
				 * nearest k/2 distances. If it is added to the top set
				 * then it is possible that a distance was dropped off
				 * the list of the top k/2, so see if that bumped distance
				 * from i -> dropped data point can be added to the dropped 
				 * data point's near list
				 */
				DEBUG2( "Added %ld to %ld's near list ...\n", j, i );

				if( BOND_DROPPED(offj) )
				{
					if( ADD_BOND_NEAR( offj, i, off_dist, off_dist2, offj2 ) )
						DEBUG2( "Added %ld to %ld's near list ...\n", offj, i );
				}
			}
			else if( ADD_BOND_FAR( i, j, ndist, off_dist, offj ) )
			{
				/*
				 * next try to add distance from i -> j to i's list
				 * of far distances. If it is added and this forces
				 * a distance to be dropped, try to add the distance
				 * from i -> dropped data point to the dropped data
				 * points far list
			     */
				DEBUG2( "Added %ld to %ld's near list ...\n", i, j );

				if( BOND_DROPPED(offj) )
				{
					if( ADD_BOND_FAR( offj, i, off_dist, off_dist2, offj2 ) )
						DEBUG2( "Added %ld to %ld's far list ...\n", i, offj );
				}
			}
			else if( ADD_BOND_NEAR( j, i, ndist, off_dist, offj ) )
			{
				/*
				 * next try to add distance from i -> j to j's list
				 * of near distances. If it is added and this forces
				 * a distance to be dropped, try to add the distance
				 * from j -> dropped data point to the dropped data
				 * points far list
			     */
				DEBUG2( "Added %ld to %ld's near list ...\n", i, j );

				if( BOND_DROPPED(offj) )
				{
					if( ADD_BOND_FAR( j, offj, off_dist, off_dist2, offj2 ) )
						DEBUG2( "Added %ld to %ld's far list ...\n", offj, j );
				}
			}
			else if( ADD_BOND_FAR( j, i, ndist, off_dist2, offj2 ) )
			{
				/*
				 * next try to add distance from i -> j to j's list
				 * of far distances. If it doesn't go there - give up
				 * we tried hard enough :-)
			     */
				DEBUG2( "Added %ld to %ld's near list ...\n", i, j );
			}
			else
				db->stats.num_dropped++;
		}
	}
	
	/*
	 * divide by total to calc mean
	 */
	n = ((A->height)*(A->height-1))/2;
	mean = sum/n;
	
	/*
	 * store resulting stats in database struct
	 */
	stats->sum_ndist = sum;
	stats->num_ndist = n;
	stats->min_ndist = min;
	stats->max_ndist = max;
	stats->mean_ndist = mean;
}

/****************************************************************/
/*	function:	add_bond_sorted() - adds a reference to a		*/
/*				2nd point to a data point's list of top bonds.	*/
/*				the ndist to the second data point is compared  */
/*				with the values already in the sorted list,		*/
/*				to determine whether any values need to be 		*/
/*				bumped to make room for the new reference		*/
/*																*/
/****************************************************************/
int add_bond_sorted( double dists[], long others[], long low, long high, 
						    long j, double ndist, long *num_bonds, 
							double *off_dist, long *offj, TST_FUNC tst )
{
	long	k; 

	DEBUG( "add_bond_sorted" );

	/*
	 * if list is empty just put info in
	 * zeroth slots of arrays
	 */
	if( *num_bonds == low )
	{
		dists[low] = ndist;
		others[low] = j;	
		(*num_bonds)++;
		return( TRUE );				
	}
	
	/*
	 * k is now the number of bonds, but is clamped
	 * to be no greater than t, and no less than 0
	 * the loop below will decrement k immediately
	 * so that k will then index the last info
	 * previously added to the arrays
	 */
	k = MIN( *num_bonds, high );

	/*
	 * search down list for ndist that's less than
	 * new ndist, shifting values to the right as we go
	 * to make room for new ndist when we find its place
	 */
	while( --k >= low && (*tst)(ndist,dists[k]) )
	{
		if( k+1 == high )
		{
			if(off_dist) 
			{
				*off_dist = dists[k];
				*offj = others[k];
			}
		}
		else
		{
			dists[k+1] = dists[k];
			others[k+1] = others[k];
		}
	}
	
	/*
	 * the loop above can terminate for two reasons:
	 * A) if k < 0, then there is no ndist in the list
	 * closer than the one we're inserting, so the new
	 * ndist should be at the head of the list (in slot 0).
	 * B) if the new ndist is farther than all the values 
	 * in the list, in which case we want to ignore the new
	 * value. We increment "k" to correct for the extra "k--" 
	 * that can terminate the above loop
	 */
	if( ++k < high )
	{
		dists[k] = ndist;
		others[k] = j;
		*num_bonds += (*num_bonds < high) ? 1 : 0;
		return( TRUE );				
	}

	return( FALSE );
}

void calc_all_sdist( DB *db, MATRIX *P, MATRIX *O, MATRIX *sD )
{
	long	i, j;
	long	other, dims;
	int		*mask;

	DEBUG( "calc_all_sdist" );

	if( !db || !P || !P->data || !O || !O->ldata || !sD || !sD->data )
		return;

	dims = db->dat.display_dims;
	mask = display_mask;
	db->stats.sum_sdist = 0.0;

	/*
	 * get distance between points in display-dimensional space
	 * also store sum of display-dimensional distances
	 */
	for( i = 0; i < P->height; i++ )
	{
		for( j = 0; j < db->dat.data[i].num_bonds; j++ )
		{
			other = O->ldata[i][j];
			sD->data[i][j] = (*distance)(P->data[other], P->data[i], mask, dims );
			db->stats.sum_sdist += sD->data[i][j];
		}
	}
}

void calc_all_stress( DB *db, MATRIX *nD, MATRIX *sD, MATRIX *S )
{
	long	i, j;
	double	stress;

	DEBUG( "calc_all_stress" );

	if( !db || !nD || !nD->data || !sD || !sD->data || !S || !S->data || 
						nD->width != sD->width || nD->height != sD->height )
		return;

	db->stats.max_datum_stress = 0;
	db->stats.max_bond_stress = 0;

	/*
	 * calculate "stress" between n-dimensional and
	 * display-dimensional representations of data
	 */
	for( i = 0; i < nD->height; i++ )
	{
		db->dat.data[i].total_stress = 0.0;

		for( j = 0; j < db->dat.data[i].num_bonds; j++ )
		{
			S->data[i][j] = (*calc_stress)( nD->data[i][j], sD->data[i][j],
									db->stats.sum_ndist, db->stats.sum_sdist );
			stress = ABS(S->data[i][j]);
			db->dat.data[i].total_stress += stress;

			db->stats.max_bond_stress = MAX(db->stats.max_bond_stress, stress);
		}

		db->stats.max_datum_stress = MAX( db->stats.max_datum_stress,
										  db->dat.data[i].total_stress);
	}

	ENSURE_NONZERO( db->stats.max_datum_stress );
	ENSURE_NONZERO( db->stats.max_bond_stress );
}

/****************************************************************/
/*	function:	calc_new_positions() - calculate a new position	*/
/*				for the display representation of a data point. */
/*				It calculates the "stress" between the display	*/
/*				coordinates and the n-dimensional coordinates	*/
/*				for every data point that "pulls" this point. 	*/
/*				The stress values are then used to scale the 	*/
/*				vector between the two points. Finally, all the	*/
/*				vectors are summed and scaled to produce an		*/
/*				overall deflection amount which is added to the */
/*				current display position and stored 			*/
/*																*/
/*	output:		none											*/	
/****************************************************************/
static void calc_new_positions( DB *db, MATRIX *PO, MATRIX *PN, MATRIX *S, 
								MATRIX *O, double step_fact, double noise_fact )
{
	long	dims, i, j, other, num_bonds;
	double	total_delta[MAX_DISPLAY_DIMS];
	double	delta[MAX_DISPLAY_DIMS];
	double	step_delta[MAX_DISPLAY_DIMS];
	double	noise_delta[MAX_DISPLAY_DIMS];
	double	step_scale;

	DEBUG( "calc_new_positions" );

	dims = db->dat.display_dims;

	/*
	 * go through list of other data points that pull 
	 * this point and calculate "stress" for each
	 */
	for( i = 0; i < PO->height; i ++ )
	{
		num_bonds = db->dat.data[i].num_bonds;
		if( num_bonds == 0 )
			continue;

		init_nvect( total_delta, dims );

		for( j = 0; j < num_bonds; j++)
		{
			/*
			 * get delta between display coords and scale this by the
			 * stress between the two points & add this to a total 
			 * displacement vector
			 */
			other = O->ldata[i][j];

			sub_nvect( PO->data[other], PO->data[i], dims, delta );
			mult_nvect( S->data[i][j], delta, dims, delta );
			add_nvect( delta, total_delta, dims, total_delta );
		}
		
		/*
		 * divide by number of bonds to 'normalize'
		 */
		step_scale = 1.0/num_bonds * step_fact;
		mult_nvect( step_scale, total_delta, dims, step_delta );

		rand_nvect( noise_delta, dims );
		mult_nvect( noise_fact, noise_delta, dims, noise_delta );

		add_nvect( step_delta, noise_delta, dims, total_delta );
	
		/*
		 * add displacement to current display coords
		 */
		add_nvect( total_delta, PO->data[i], dims, PN->data[i] );
	}
}

int is_nearer( double new_dist, double old_dist )
{
	DEBUG( "is_nearer" );

	return( new_dist < old_dist );
}

int is_further( double new_dist, double old_dist )
{
	DEBUG( "is_further" );

	return( new_dist > old_dist );
}

static double calc_stress_lingoes(  double ndist, double sdist, 
									double ndist_sum, double sdist_sum )
{
	DEBUG( "calc_stress_lingoes" );

	return(  FLT_ZERO(sdist) ? 0.0 : (sdist - ndist)/sdist );
}

static double calc_stress_sammon( double ndist, double sdist,
								  double ndist_sum, double sdist_sum )
{
	DEBUG( "calc_stress_sammon" );

	return(FLT_ZERO(ndist*ndist_sum)? 0.0 : SQR(ndist-sdist)/(ndist*ndist_sum));
}

static double calc_stress_J_ee( double ndist, double sdist, 
								double ndist_sum, double sdist_sum )
{
	DEBUG( "calc_stress_J_ee" );

	return( FLT_ZERO(ndist_sum)? 0.0 : SQR(sdist - ndist)/SQR(ndist_sum) );
}

static double calc_stress_J_ff( double ndist, double sdist, 
								double ndist_sum, double sdist_sum )
{
	DEBUG( "calc_stress_J_ff" );

	return(  FLT_ZERO(ndist) ? 0.0 : SQR((sdist - ndist)/ndist) );
}

static double calc_stress_J_ef( double ndist, double sdist, 
								double ndist_sum, double sdist_sum )
{
	DEBUG( "calc_stress_J_ef" );

	return(FLT_ZERO(ndist*ndist_sum) ? 0.0: SQR(sdist-ndist)/(ndist*ndist_sum));
}

/****************************************************************/
/*	function	calc_all_ndist_random() - for every point in 	*/
/*				calculate its distance to every other point. 	*/
/*				Then for each point save the top NUM_ATTRACTION */
/*				points which exert the greatest effect on it.  	*/
/*				Also calculate min, max, mean, standard dev		*/
/*				of n-dimensional distances,	and save these 		*/
/*				statistics in the database structure			*/
/*																*/
/*	input:		db 		- pointer to database					*/
/*	output:		none											*/
/****************************************************************/
static void calc_all_ndist_random( DB *db, MATRIX *A, MATRIX *nD, MATRIX *O, 
								   long limit, int mask[], MDS_STATS *stats )
{
	long		i, j, k, n = 0;
	double		ndist; 
	double		min = DBL_MAX;
	double		max = DBL_MIN;
	double		sum = 0.0;
	
	DEBUG( "calc_all_ndist_random" );

	/*
	 * calculate all n(n-1)/2 distances between data points
	 * also calculate mean, min, max n-dimensional distances
	 */	
	for( i = 0; i < nD->height; i++ )
	{
		for( k = 0; k < MIN(limit,i); k++ )
		{
			if( k == 0 )
				ndist = min_distance_from_i( A, mask, i, &j );
			else if( k == 1 )
				ndist = max_distance_from_i( A, mask, i, &j );
			else
			{
				j = irand( 0, i );
				
				ndist = (*distance)( A->data[i], A->data[j], mask, A->width);
			}

			min = MIN( ndist, min );
			max = MAX( ndist, max );
			sum += ndist;

			nD->data[i][k] = ndist;
			O->ldata[i][k] = j;

			n++;
		}

		db->dat.data[i].num_bonds = k;
	}

    /*
     * store resulting stats in database struct
     */
    stats->sum_ndist = sum;
    stats->num_ndist = ((A->height)*(A->height-1))/2;
    stats->min_ndist = min;
    stats->max_ndist = max;
    stats->mean_ndist = sum/n;
}
