/****************************************************************/
/* headers                                                      */
/****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "general.h"
#include "util.h"
#include "nmatrix.h"
#include "viz.h"
#include "bonds.h"
#include "nvect.h"
#include "cluster.h"
#include "db.h"
	
/****************************************************************/
/* defines                                                      */
/****************************************************************/

/****************************************************************/
/* typedefs                                                     */
/****************************************************************/

/****************************************************************/
/* external function                                            */
/****************************************************************/
/****************************************************************/
/* function prototypes                                          */
/****************************************************************/
/*
 * operations on entire database
 */
DB *alloc_db();
int alloc_data( DB *db );
int allocate_data_matrices( DB *db );
int allocate_display_matrices( DB *db );
int allocate_subset_matrices( DB *db );
void init_db( DB * db );
int init_dim_mask( DB *db, int len );
void free_db( DB * );
void free_data( DB *db );
void init_rand( DB *db );

/****************************************************************/
/* global variables												*/
/****************************************************************/
extern char errbuf[];

/****************************************************************/
/*	function:	alloc_db() - allocate memory for a database		*/
/*				containing datums, each of which has a n-vector	*/
/*				of values associated with it					*/
/*																*/
/*	input:		none											*/	
/*																*/
/*	output:		pointer to allocated database struct			*/	
/****************************************************************/
DB *alloc_db()
{
	DB		*db;
	
	DEBUG( "alloc_db" );

	/*
	 * allocate database struct
	 */
	if( (db = (DB *)mmalloc( sizeof(DB) )) == NULL )
		return( NULL );
	
	/*
	 * set database to defaults
	 */
	init_db( db );

	return( db );
}

/****************************************************************/
/*	function:	alloc_data() - allocate memory to hold actual	*/
/*				data points or "datums", each of which is an	*/
/*				n-dimensional vector of values plus a few other	*/
/*				fields for bookkeeping							*/
/*																*/
/*	input:		pointer to database struct						*/	
/*																*/
/*	output:		ERROR if the memory allocation fails, 			*/
/*				otherwise SUCCESS								*/
/****************************************************************/
int alloc_data( DB *db )
{
	DEBUG( "alloc_data" );

	if( !db )
		return( ERROR );

	if( allocate_data_matrices( db ) == ERROR ||
		allocate_display_matrices( db ) == ERROR ||
		allocate_subset_matrices( db ) == ERROR )
	{
		free_db( db );
		return( ERROR );
	}

	return( SUCCESS );
}

int allocate_data_matrices( DB *db )
{
	long	data_dims, m;

	DEBUG( "allocate_data_matrices" );

	if( !db )
		return ERROR;

	data_dims = db->dat.data_dims;
	m = db->dat.num_points;
	db->dat.num_dists = (m*(m-1))/2;

	if ( db->dat.A ) free_matrix_mem( db->dat.A );
	if ( db->dat.R ) free_matrix_mem( db->dat.R );

	/*
	 * allocate matrix to hold unnormalized n-dimensional data points
	 * allocate matrix to hold normalized n-dimensional data points
	 */
	if( !(db->dat.A = alloc_matrix(data_dims, m, DOUBLE_TYPE)) ||
		!(db->dat.R = alloc_matrix(data_dims, m, DOUBLE_TYPE)) )
	{
		db->dat.A = NULL;
		db->dat.R = NULL;
		return( ERROR );
	}

	/*
	 * allocate actual datums
	 */	
	if( !(db->dat.data = (DATUM *)mcalloc(db->dat.num_points, sizeof(DATUM))))
		return( ERROR );

	/*
	 * initialize distance function masks
	 */
	if( init_dim_mask( db, data_dims ) == ERROR )
		return( ERROR );

	/*
	 * allocate dim info array
	 */
	db->dat.data_dim_array = (DIM_INFO *)mmalloc(data_dims*sizeof(DIM_INFO));
	if( db->dat.data_dim_array == NULL )
        return( ERROR );


	return( SUCCESS );
}

int allocate_subset_matrices( DB *db )
{
	long	m;

	DEBUG( "allocate_subset_matrices" );

	if( !db )
		return ERROR;

	m = db->dat.num_points;

	free_matrix_mem( db->dat.nD );
	free_matrix_mem( db->dat.sD );
	free_matrix_mem( db->dat.O );
	free_matrix_mem( db->dat.S );

	/*
	 * allocate matrix to hold k distances between n-dimensional data points
	 * allocate matrix to hold k distances between display points
	 * allocate matrix to hold k references to other points
	 * allocate matrix to hold k stresses for each data points
	 */
	if( !(db->dat.nD = alloc_matrix(db->mds.num_bonds, m, DOUBLE_TYPE)) ||
		!(db->dat.sD = alloc_matrix(db->mds.num_bonds, m, DOUBLE_TYPE)) ||
		!(db->dat.O = alloc_matrix(db->mds.num_bonds, m, LONG_TYPE)) ||
		!(db->dat.S = alloc_matrix(db->mds.num_bonds, m, DOUBLE_TYPE)) )
	{
		db->dat.nD = NULL;
		db->dat.sD = NULL;
		db->dat.O = NULL;
		db->dat.S = NULL;
		return( ERROR );
	}

	return( SUCCESS );
}

int allocate_display_matrices( DB *db )
{
	long	display_dims, m;

	DEBUG( "allocate_display_matrices" );

	if( !db )
		return ERROR;

	display_dims = db->dat.display_dims;
	m = db->dat.num_points;

	free_matrix_mem( db->dat.P );
	free_matrix_mem( db->dat.PO );
	free_matrix_mem( db->dat.POO );
	free_matrix_mem( db->dat.POOO );
	free_matrix_mem( db->dat.IP );

	/*
	 * allocate matrix to hold p-dimensional display points
	 * allocate matrices to hold previous iteration's p-dimensional display pnts
	 * allocate matrix to hold interpolated display points
	 */
	if( !(db->dat.P = alloc_matrix(display_dims, m, DOUBLE_TYPE)) ||
		!(db->dat.PO = alloc_matrix(display_dims, m, DOUBLE_TYPE)) ||
		!(db->dat.POO = alloc_matrix(display_dims, m, DOUBLE_TYPE)) ||
		!(db->dat.POOO = alloc_matrix(display_dims, m, DOUBLE_TYPE)) ||
		!(db->dat.IP = alloc_matrix(display_dims, m, DOUBLE_TYPE)) )
	{
		db->dat.P = NULL;
		db->dat.PO = NULL;
		db->dat.POO = NULL;
		db->dat.POOO = NULL;
		db->dat.IP = NULL;

		return( ERROR );
	}

	if( init_display_mask( db->dat.display_dims ) == ERROR )
		return( ERROR );

	return( SUCCESS );
}

/****************************************************************/
/*	function:	free_db() - deallocate memory for a database	*/
/*				containing datums, each of which has a n-vector	*/
/*				of values associated with it. Frees memory		*/
/*				used by datums, as well as that used by 		*/
/*				heirarchy of cluster structs (which are really	*/
/*				also datums)									*/
/*																*/
/*	input:		pointer to database to free						*/
/*																*/
/*	output:		none											*/	
/****************************************************************/
void free_db( DB *db )
{
	DEBUG( "free_db" );

	if( !db )
		return;
	
	/*
	 * free cluster tree and datum array
	 */
	free_data( db );

	/*
	 * close report file
	 */
	if( db->output_mode & TEXT && db->files.reportfp )
		fclose( db->files.reportfp );

	mfree( db->name );
	mfree( db->files.infile );
	mfree( db->files.outfile );
	mfree( db->files.viewfile );
	mfree( db->animate.frame_base );

	/*
	 * free memory used by db struct
	 */
	mfree( db );
}

/****************************************************************/
/*	function:	free_data() - frees memory used to hold data 	*/
/*				points or "datums" in the database				*/
/*																*/
/*	input:		pointer to database struct						*/	
/*																*/
/*	output:		none											*/
/****************************************************************/
void free_data( DB *db )
{
	long			i;
	NOMINAL_INFO	*ptr, *delptr;

	DEBUG( "free_data" );

	if( !db || !db->dat.data )
		return;

	/*
	 * free allocated fields of datums
	 */	
	for( i = 0; i <  db->dat.num_points; i++ )
	{
		/*
		 * free datum label
		 */
		mfree( db->dat.data[i].label );

		/*
		 * free datum's linked list of nominals
		 */
		ptr = db->dat.data[i].nominal_head;
		while( ptr )
		{
			delptr = ptr;
			mfree( delptr->name );
			ptr = ptr->next;

			mfree( delptr );
		}
	}
	
	/*
	 * free array of datums and matrices holding
	 * n-dimensional data points, display points and
	 * previous iteration display points
	 */
	free_matrix_mem( db->dat.A );
	free_matrix_mem( db->dat.P );
	free_matrix_mem( db->dat.PO );
	free_matrix_mem( db->dat.POO );
	free_matrix_mem( db->dat.POOO );
	free_matrix_mem( db->dat.IP );
	free_matrix_mem( db->dat.nD );
	free_matrix_mem( db->dat.sD );
	free_matrix_mem( db->dat.S );
	free_matrix_mem( db->dat.O );
	mfree( db->dat.data );

    db->dat.A = NULL;
    db->dat.P = NULL;
    db->dat.PO = NULL;
    db->dat.POO = NULL;
    db->dat.POOO = NULL;
    db->dat.IP = NULL;
    db->dat.nD = NULL;
    db->dat.sD = NULL;
    db->dat.S = NULL;
    db->dat.O = NULL;
	db->dat.data = NULL;
}

/****************************************************************/
/*	function:	init_data() - initializes a database to have 	*/
/*				reasonable default values						*/
/*																*/
/*	input:		pointer to database struct						*/	
/*																*/
/*	output:		none											*/
/****************************************************************/
void init_db( DB *db )
{
	DEBUG( "init_data" );

	if( !db )
		return;

	/*
	 * init database fields to defaults
	 */
	db->name = NULL;
	db->interact_mode = ROTATE;
	db->output_mode = TEXT;
	db->draw_mode = ATTRACTIONS;
	db->plot_mode = 0;

	db->files.infile = mstrdup( "data\\teapot.dat" );
	db->files.outfile = mstrdup( "out.mds" );
	db->files.viewfile = NULL;
	db->files.reportfp = NULL;

	db->animate.max_iteration = DEFAULT_MAX_ITER;
	db->animate.save_frames = FALSE;
	db->animate.interpolate = FALSE;
	db->animate.interp_steps = 4;
	db->animate.frame_base = mstrdup( "frame" );
	db->animate.start_save_iter = 0;
	db->animate.end_save_iter = DEFAULT_MAX_ITER;
	db->animate.save_skip = 1;

	db->dat.num_clusters = DEFAULT_NUM_POINTS;
	db->dat.num_points = DEFAULT_NUM_POINTS;
	db->dat.num_dists = db->dat.num_points * (db->dat.num_points - 1)/2;
	db->dat.data_dims = DEFAULT_DATA_DIMS;
	db->dat.display_dims = DEFAULT_DISPLAY_DIMS;
	db->dat.data_init_mode = FROM_FILE;
	db->dat.num_P = 0;
	db->dat.clust_prob = 1.0;
	db->dat.seed = time((time_t *)0);
    db->dat.A = NULL;
    db->dat.R = NULL;
    db->dat.P = NULL;
    db->dat.PO = NULL;
    db->dat.POO = NULL;
    db->dat.POOO = NULL;
    db->dat.IP = NULL;
    db->dat.nD = NULL;
    db->dat.sD = NULL;
    db->dat.S = NULL;
    db->dat.O = NULL;
	db->dat.data = NULL;

	db->mds.display_init_mode = RANDOM;
	db->mds.select_mode = TOP_TEN;
	db->mds.dist_func = EUCLIDIAN;
	db->mds.step_factor = 1.0;
    db->mds.noise_factor = 0.0;
	db->mds.stress_function = LINGOES_STRESS;
	set_stress_func( db );
	db->mds.norm_function = UNIT_VARIANCE;
	set_normalize( db->mds.norm_function );
	db->mds.num_bonds = DEFAULT_NUM_ATTRACTORS;

	db->stats.num_ndist = 0;
	db->stats.min_ndist = 0.0;
	db->stats.max_ndist = 0.0;
	db->stats.mean_ndist = 0.0;
	db->stats.max_datum_stress = 1.0;
	db->stats.max_bond_stress = 1.0;
	db->stats.start_max_datum_stress = 1.0;
	db->stats.start_max_bond_stress = 1.0;

	/*
	 * initialize error function pointer
	 */
	error = error_tty;
}

/****************************************************************/
/*	function:	init_dim_mask() - initializes the mask that		*/
/*				determines whether a given dimension is 		*/
/*				included in the distance calculations that are	*/
/*				the heart of the MDS procedure. The default 	*/
/*				state is for all the dimensions to be "enabled"	*/
/*				i.e. to have '1's in every cell of the mask		*/
/*																*/
/*	input:		db	- pointer to database struct				*/	
/*				len	- length of mask							*/	
/*																*/
/*	output:		SUCCESS if mask allocated, ERROR otherwise		*/
/****************************************************************/
int init_dim_mask( DB *db, int len )
{
	int		i;

	DEBUG( "init_dim_mask" );

	if( (db->dat.dim_mask = (int *)mmalloc(len*sizeof(int))) == NULL )
        return( ERROR );

	for( i = 0; i < len; i++ )
		db->dat.dim_mask[i] = 1;

	return( SUCCESS );
}

/****************************************************************/
/*	function:	init_rand() - sets the seed of the random		*/
/*				number generator to the seed value stored in	*/
/*				the database struct. By saving the seed here	*/
/*				we can "replay" an animation of MDS by just		*/
/*				restoring the old seed and rerunning			*/
/*																*/
/*	input:		pointer to database struct						*/	
/*																*/
/*	output:		none											*/
/****************************************************************/
void init_rand( DB *db )
{
	DEBUG( "init_rand" );

//    srand( db->dat.seed );
}
