#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <search.h>


/*************************************************************************/
//Code for glyph shift

/* program to adjust the positions of glyphs to reduce or eliminate overlaps.
Overlapping points exert a force at each other in the direction of the
vector between them, weighted by how close they are to each other.  Multiple
overlaps are simply summed and averaged.  Written by Matt Ward, last changed 8/2/00
*/

#define MAX_ITER 10             /* maximum number of iterations of adjustment(change as needed) */
#define SIZE 25                 /* maximum size of a glyph in x or y (change as needed) */
#define TOLERANCE 2             /* maximum allowable percentage of glyph overlaps (can be changed) */
#define SPEED SIZE/4            /* maximum change in position per iteration (arbitrary)*/

/* macro to determine if both x and y extents overlap between 2 glyphs - may have it backwards */       
#define OVER(a, b)  (((abs(pos[a].x-pos[b].x) < SIZE) && (abs(pos[a].y-pos[b].y) < SIZE))? 1:0)

struct position {               /* structure to hold a position */
   int x, y;
   };

struct position *pos;           /* buffer to hold glyph positions */
struct position *tpos;          /* temporary buffer for positions */

int num_pts;                    /* number of data points */

/* initialize space for positions and store them */
void init(int pts, double *basex, double *basey, double para)
{
	num_pts = pts;
	pos = (struct position *)malloc(pts * sizeof(struct position));
	tpos = (struct position *)malloc(pts * sizeof(struct position));
	if(pos == NULL || tpos == NULL)
	{
	  printf("Malloc failed in init\n");
	  exit(-1);
	}
/* now fill in positions of glyphs, whether data driven or derived */
/* CODE TO INITIALIZE POSITIONS GOES HERE */
	for ( int i = 0; i < pts; i++ )
	{
		pos[i].x = tpos[i].x = int ( 0.5 + basex[i]*para);
		pos[i].y = tpos[i].y = int ( 0.5 + basey[i]*para);
	}
}

/* count number of overlaps based on overlapping extents */
int num_overlaps(void)
{
   int i, j, num = 0;
   for(i = 0;i < num_pts;i++)
      for(j = i+1;j < num_pts;j++)
         num += OVER(i, j);
   return(num);
}

/* need to determine which points to move, what direction, and how much.
Points to move are those that overlap.  The direction of motion for point i is the
combined forces from all points that overlap it, formed as the vectors going
from the overlapping point to point i.  The vectors are weighted, so closer
points have more influence on the direction.  The overall motion is weighted by
SPEED, the maximum number of pixels a glyph can move */
void move_glyphs(void)
{
	int i, j, num;
	double mag;
	double vector[2], sumvector[2], maxsize;
	maxsize = (double)SIZE * 1.415;       /* max distance between overlap centers */
	sumvector[0] = sumvector[1] = 0.0;
	for(i = 0;i < num_pts;i++)   
	{
		num = 0;
/* for every point... */
		for(j = 0;j < num_pts;j++)        
		{
/* if it has any overlap */
			if(i != j && OVER(i, j))       
			{
				num++;
/* calculate vector between centers (the direction of force repelling points) */

/* if points are at same location, add a random displacement */
				if((pos[j].x == pos[i].x) && (pos[j].y == pos[i].y)) 
				{
					vector[0] = -1.0 + 2.0 * rand();        /* rand num between -1 and 1 */
					vector[1] = -1.0 + 2.0 * rand();
					mag = sqrt((vector[0] * vector[0]) + (vector[1] * vector[1]));
					/* normalize, and add with full weight to sumvector */
					vector[0] = vector[0]/mag;
					vector[1] = vector[1]/mag;
					sumvector[0] += vector[0];
					sumvector[1] += vector[1];
				}
/* normalize vector and add weighted amount to sumvector */
				else 
				{
					vector[0] = pos[i].x - pos[j].x;
					vector[1] = pos[i].y - pos[j].y;
					mag = sqrt((vector[0] * vector[0]) + (vector[1] * vector[1]));
					vector[0] = vector[0]/mag;
					vector[1] = vector[1]/mag;
					/* weight decreases as distance increases, from 1.0 to 0.0 */
					sumvector[0] += (1.0-mag/maxsize) * vector[0];
					sumvector[1] += (1.0-mag/maxsize) * vector[1];
				}
			}
		}
/* NOTE: may want to add a minor random offset for points being moved, to avoid
   oscillating between 2 locations */

/* average the vectors (this is not exactly correct, as lots of many nearby
points can dillute influence of very close one, resulting in smaller move, but in
correct direction)  */
		if(num > 0)       
		{
			sumvector[0] = sumvector[0]/(double)num;
			sumvector[1] = sumvector[1]/(double)num;
		}
/* calculate the new position for glyph i */
		tpos[i].x = (int)(pos[i].x + SPEED * sumvector[0]);
		tpos[i].y = (int)(pos[i].y + SPEED * sumvector[1]);
	}
/* now copy everything back into pos array */
	for(i = 0;i < num_pts;i++)   
	{
		pos[i].x = tpos[i].x;
		pos[i].y = tpos[i].y;
	}
}


/* compute tolerance as number of allowable overlaps and iteratively move
     points until tolerance met or maximum iterations reached */
void adjust(void)
{
   int i, tol;
   tol = TOLERANCE * num_pts / 100; 
   for(i = 0;i < MAX_ITER && num_overlaps() > tol;i++)
      move_glyphs();
}

void finished(int pts, double *basex, double *basey, double para)
{
//put back the glyph basex and basey
	for ( int i = 0; i < pts; i++ )
	{
		basex[i] = pos[i].x/para;
		basey[i] = pos[i].y/para;
	}
}

void Shift(int data_size, float* pos1, float* pos2, float para)
{
	init(data_size, pos1, pos2, para);	
	adjust();
	finished(data_size, pos1, pos2, para);
}
