#ifndef GRCANVAS_H
#define GRCANVAS_H

#include <wx/glcanvas.h>
#include <wx/hashmap.h>
#include <GL/glu.h>
#include <string>
#include "../Basic/Vec2.h"
#include "../Basic/Vec3.h"
#include "../Basic/RGB.h"
#include "../Basic/ARR.h"
#include "../Basic/macros.h"

#define VAR_HASH 0
#define GLYPH_HASH 1
#define DIMSTACK_HASH 2
#define TIME_HISTOGRAM_HASH 3

class ViewManager;
class GLFrame;

//structure grTexture contains the gl id and shape of a texture oject created by OpenGL
struct grTexture
{
    GLuint id;
    //width and height are the size of the texture object
    int width;
    int height;
    //width1 and height are the actual size of the figure contained in the texture object
    int width1;
    int height1;

    grTexture(GLuint gid, int gwidth, int gheight, int gwidth1, int gheight1) 
		: id(gid), width(gwidth), height(gheight), width1(gwidth1), height1(gheight1) {}

    grTexture() 
		: id(NULL), width(0), height(0), width1(0), height1(0) {}
};


WX_DECLARE_STRING_HASH_MAP(grTexture, TextHash);
WX_DECLARE_HASH_MAP(int, grTexture, wxIntegerHash, wxIntegerEqual, ITHash);


//canvas class defined by the system that allows the usage of all opengl command
//this class also privdes functions for drawing common shapes such as lines, circles, and others
//it keeps tracks of the texture objects created
class grCanvas : public wxGLCanvas
{
protected:
    ViewManager* viewManager;

	Vec2 zoom; //zoom factors of the canvas
    Vec2 translate; //pan factors of the canvas
    bool origin_at_lower_left;  //origin of the canvas

    int viewPortType;

	RGBt fg_color;
    RGBt bg_color;
    double alpha;
    bool translucent;
    GLbitfield clear_bits;
	bool xxor;

    bool shiftkey;
    bool controlkey;
    bool altkey;
    bool capitalkey;

    TextHash textHash;

    ITHash varHash;
    ITHash glyphHash;
    ITHash dimstackHash;
    ITHash timeHistogramHash;
    //This array contains the above hash tables. They can be accessed though this 
    //this array according to their id defined in this header file
    //i.e., VAR_TEXTURE_HASH_ID, GLYPH_TEXTURE_HASH_ID, DIMSTACK_TEXTURE_HASH_ID, and TIME_HISTOGRAM_HASH_ID
    ITHash textureHashs[4];

    GLFrame* frame;

protected:
    void SetViewportInt();
    void SetupTransformMatrix();

    void CreateTextTexture(const char* str, RGBt bgColor, RGBt fgColor, GLuint &texture, 
        int &width, int &height, int &width1, int &height1);
    void GetTextTexture(std::string str, GLuint &texture, int &width, int &height, int &width1, int &height1);
    void DrawTextTexture(GLuint texture, int width, int height, int width1, int height1, int leftX, int bottomY, 
                            float angle, float scale);
  
    void CreateTexture(int width, int height, float* data, GLuint &texture);
    void DeleteTextTextures();

    void BeginDraw();
    void EndDraw();

    Vec2 GetMousePositionInDataSpace(wxMouseEvent event);

    ////////////// event handlers //////////////////////
	void OnSize(wxSizeEvent& event);
	
    void OnPaint(wxPaintEvent& event);

	//Process a wxEVT_LEFT_DOWN event.  
	void OnLeftButtonDown(wxMouseEvent &event); 

	//Process a wxEVT_LEFT_UP event.  
	void OnLeftButtonUp(wxMouseEvent &event); 

	//Process a wxEVT_LEFT_DCLICK event.
	void OnLeftButtonDoubleClick(wxMouseEvent &event); 

	//Process a wxEVT_MIDDLE_DOWN event. 
	void OnMiddleButtonDown(wxMouseEvent &event); 

	//Process a wxEVT_MIDDLE_UP event. 
	void OnMiddleButtonUp(wxMouseEvent &event); 

	//Process a wxEVT_MIDDLE_DCLICK event. 
	void OnMiddleButtonDoubleClick(wxMouseEvent &event); 

	//Process a wxEVT_MIDDLE_DCLICK event. 
	void OnRightButtonDown(wxMouseEvent &event); 

	//Process a wxEVT_RIGHT_UP event. 
	void OnRightButtonUp(wxMouseEvent &event); 

	//Process a wxEVT_RIGHT_DCLICK event. 
	void OnRightButtonDoubleClick(wxMouseEvent &event);

	//Process a wxEVT_MOTION event. 
	void OnMouseMotion(wxMouseEvent &event); 

    void OnKeyPress(wxKeyEvent &event);
    void OnKeyRelease(wxKeyEvent &event);

	////////////// event handlers //////////////////////
    void CheckAltShiftControl(wxMouseEvent &event);


public:
	grCanvas(wxWindow* parent, wxWindowID id, const wxPoint& pos, 
		     const wxSize& size, long style=0, const wxString& name="GLCanvas");
	~grCanvas();

    void SetViewManager(ViewManager* viewManager);
    void SetFrame(GLFrame* frame) {this->frame = frame;}

    void SetViewport();
    void Refresh();
    
    //write a message in the status bar of the glFrame
    //i can be 0, 1, or 2, since there are 3 message place in the frame status bar
    void WriteMessage(std::string message, int i);
    //this file create a new image file that contains the content in the current display of the canvas
    void WriteImage(std::string fileName);

    Vec2 MapScreenToData( const Vec2 &sc );
    Vec2 MapDataToScreen( const Vec2 &p );
    Vec2 MapScreeVectorToDataVector(const Vec2 p);

    void DrawString(std::string str, float baseX, float baseY, float angle, float scale);
    float GetFontHeight();

    void SetLowerLeftOrigin();
    void SetTopLeftOrigin();

    //set and get the translate (panning) and zooming factors
    void SetXTranslate(float value);
    void SetYTranslate(float value);
    float GetXTranslate();
    float GetYTranslate();
    void SetXZoom(float value);
    void SetYZoom(float value);
    float GetXZoom();
    float GetYZoom();

    void SetForeground( const RGBt &color);
    void SetBackground( const RGBt &color);
    void SetLineWidth(float width);
    void Clear();

    void DrawLine( double x1, double y1, double x2, double y2);
    void DrawCircle( double centerx, double centery, double radius, double offset);
    void DrawArc( double centerx, double centery, double radius, double start_angle, double sweep_angle);
    void DrawPartialDisk( double centerx, double centery, double inner_radius, 
					      double outer_radius, double start_angle, double sweep_angle);
    void DrawLines( const ARR<Vec2> &pts );
    void DrawRectangle(double x, double y, double width, double ht );
    void DrawRectangle2(double x1, double x2, double y1, double y2 );
    void FillRectangle( double x, double y, double width, double ht);
    void DrawPoint( double x, double y, double size);
    void DrawPoints( const ARR<Vec2> &pts, double size);
    void DrawPolygon( const ARR<Vec2> &pts);
    void FillPolygon( const ARR<Vec2> &pts);
    void DrawTransparentBand( const ARR<Vec2> &pts,
                              double opacity_bottom, double opacity_top,
                              RGBt color_bottom, RGBt color_top);
    void SetHandCursor();
    void SetBullseyeCursor();
    void SetDefaultCursor();

    void CreateTexture(int groupID, int glyphID, int width, int height, float* data, GLuint &texture);
    bool GetTexture(int groupID, int glyphID, GLuint &texture, int &width, int &height);
    void DeleteTexture(int groupID, int glyphID);
    void DeleteTextures(int groupID);

    void KeyPress(int key);
    void KeyRelease(int key);

    bool IsShiftKey() {return shiftkey;}
    bool IsAltKey() {return altkey;}
    bool IsControlKey() {return controlkey;}
    bool IsCapitalKey() {return capitalkey;}

    Vec3 MapDataToScreen( const Vec3 &p );
    void DrawString(std::string str, float baseX, float baseY, float z, float angle, float scale);
    void DrawTextTexture(GLuint texture, int width, int height, int width1, int height1,
                               float leftX, float bottomY, float z, float angle, float scale);

    void SetTranslucent(bool trans);
    void SetViewPortMode(int type);

	DECLARE_EVENT_TABLE()

};


#endif

