#include "ViewManager.h"
#include "Display/Var/VarDrawer.h"
#include "Display/Parcoord/ParcoordDrawer.h"
#include "Display/Glyph/GlyphDrawer.h"
#include "Display/Dimstack/DimstackDrawer.h"

ViewManager::ViewManager()
{
    colorManager = new ColorManager();
}


ViewManager::~ViewManager()
{
    //delete drawers
    IPHash::iterator it;
    it = typeDrawerHash.begin();
    while (it != typeDrawerHash.end())
    {
        delete (Drawer *)(it->second);
        it ++;
    }

    //delete colorManager
    if (colorManager != NULL)
        delete colorManager;
}


///////  drawer management functions   /////////////////////////

//get the drawer assigned to a canvas
Drawer* ViewManager::GetDrawer(grCanvas* canvas)
{
    return (Drawer*) canvasDrawerHash[canvas]; 
}


//assign a drawer to a canvas according to the drawing type
//if a drawer of the required type has been created already, then connect it to the canvas
//otherwise create a drawer of the required type, then connect it to the canvas
//if the canvas was assigned a different drawer before, disconnect the drawer with the canvas
void ViewManager::AssignDrawerToCanvas(grCanvas* canvas, int type)
{
    canvas->SetViewManager(this);

    //find the drawer that is connected with canvas now (old drawer) by looking up the canvasDrawerHash
    Drawer* oldDrawer = (Drawer*) canvasDrawerHash[canvas];
    //get a drawer with the required type (new drawer) from the drawer list of the viewManager
    Drawer* newDrawer = (Drawer*) CreateDrawer(type);

    //if they are the same drawer, then do nothing
    if (oldDrawer == newDrawer)
        return;

    //disconnect the old drawer with the canvas
    if (oldDrawer != NULL)
        oldDrawer->RemoveCanvas(canvas);

    //connect the new drawer with the canvas
    canvasDrawerHash[canvas] = newDrawer;
    if (newDrawer != NULL)
        newDrawer->AddCanvas(canvas);
}


Drawer* ViewManager::GetDrawer(int type)
{
    return (Drawer*) typeDrawerHash[type];
}

//get a drawer of required drawer type
//if the drawer of required type has been created before, it is returned
//otherwise a new drawer of the required type is created and returned
//So the viewmanager will only have at most one drawer for each type. 
//to do: allow more than one drawer for each type
Drawer* ViewManager::CreateDrawer(int type)
{
    Drawer* drawer = (Drawer*) typeDrawerHash[type];
    if (drawer == NULL)
    {
        if (type == NODISPLAY)
        {
            drawer = NULL;
        }
        else if (type == VAR)
        {
            drawer = new VarDrawer();
        }
        else if (type == PARCOORD)
        {
            drawer = new ParcoordDrawer();
        }
        else if (type == GLYPH)
        {
            drawer = new GlyphDrawer();
        }
        else if (type == DIMSTACK)
        {
            drawer = new DimstackDrawer();
        }
            
        if (drawer != NULL)
        {
            drawer->SetViewManager(this);
 
        }
        typeDrawerHash[type] = drawer;
    }
    return (Drawer*) drawer;
}


///////////  canvas refreshing functions. These functions are called when the system wants to referesh one or more canvases  ////////////////

//refresh the display drawn on a canvas
void ViewManager::RefreshCanvas(grCanvas* canvas)
{
    if (docManager->IsEmpty())
        return; 
    Drawer* drawer = GetDrawer(canvas);
    drawer->RefreshCanvas(canvas);
}


//refresh the displays drawn on all canvases
void ViewManager::RefreshAll()
{
    IPHash::iterator it;
    it = typeDrawerHash.begin();
    Drawer* drawer;
    while (it != typeDrawerHash.end())
    {
        drawer = (Drawer *)(it->second);
        drawer->Refresh();
        it ++;
    }
}


///////////  canvas event handling functions  ////////////////

void ViewManager::KeyPress(grCanvas* canvas, int key)
{
    if (docManager->IsEmpty())
        return; 
    Drawer* drawer = GetDrawer(canvas);
    drawer->KeyPress(canvas, key);
}


void ViewManager::KeyRelease(grCanvas* canvas, int key)
{
    if (docManager->IsEmpty())
        return; 
    Drawer* drawer = GetDrawer(canvas);
    drawer->KeyRelease(canvas, key);
}


void ViewManager::ButtonDown(grCanvas* canvas, int button, Vec2 p) 
{
    if (docManager->IsEmpty())
        return; 
    Drawer* drawer = GetDrawer(canvas);
    drawer->ButtonDown(canvas, button, p);
}


void ViewManager::ButtonUp(grCanvas* canvas, int button, Vec2 p) 
{
    if (docManager->IsEmpty())
        return; 
    Drawer* drawer = GetDrawer(canvas);
    drawer->ButtonUp(canvas, button, p);
}


void ViewManager::DoubleClick(grCanvas* canvas, int button, Vec2 p) 
{
    if (docManager->IsEmpty())
        return; 
    Drawer* drawer = GetDrawer(canvas);
    drawer->DoubleClick(canvas, button, p);
}

void ViewManager::MouseMotion(grCanvas* canvas, Vec2 p) 
{
    if (docManager->IsEmpty())
        return; 
    Drawer* drawer = GetDrawer(canvas);
    drawer->MouseMotion(canvas, p);
}

void ViewManager::AllDimensionSpaceDeleted()
{
    IPHash::iterator it;
    it = typeDrawerHash.begin();
    Drawer* drawer;
    while (it != typeDrawerHash.end())
    {
        drawer = (Drawer *)(it->second);
        drawer->AllDimensionSpaceDeleted();
        it ++;
    }
}

void ViewManager::CurrentDimensionSpaceChanged()
{
    IPHash::iterator it;
    it = typeDrawerHash.begin();
    Drawer* drawer;
    while (it != typeDrawerHash.end())
    {
        drawer = (Drawer *)(it->second);
        if (drawer != NULL)
            drawer->CurrentDimensionSpaceChanged();
        it ++;
    }
}

void ViewManager::VarSizeChanged(float value)
{
    VarDrawer *drawer = (VarDrawer*)GetDrawer(VAR);
    if (drawer != NULL)
    {
        VarDrawerSettings *settings = drawer->GetSettings();
        settings->ChangeGlyphSize(value);
        drawer->Refresh();
    }
}

void ViewManager::GlyphSizeChanged(float value)
{
    GlyphDrawer *drawer = (GlyphDrawer*)GetDrawer(GLYPH);
    if (drawer != NULL)
    {
        GlyphDrawerSettings *settings = drawer->GetSettings();
        settings->ChangeGlyphSize(value);
        drawer->Refresh();
    }
}

void ViewManager::GlyphDistortionChanged(float value)
{
    GlyphDrawer *drawer = (GlyphDrawer*)GetDrawer(GLYPH);
    if (drawer != NULL)
    {
        GlyphDrawerSettings *settings = drawer->GetSettings();
        settings->ChangeDistortionFactor(value);
        drawer->Refresh();
    }
}

void ViewManager::VarDistortionChanged(float value)
{
    VarDrawer *drawer = (VarDrawer*)GetDrawer(VAR);
    if (drawer != NULL)
    {
        VarDrawerSettings *settings = drawer->GetSettings();
        settings->ChangeDistortionFactor(value);
        drawer->Refresh();
    }
}

void ViewManager::VarMaskColorChanged(RGBt color)
{
    VarDrawer *drawer = (VarDrawer*)GetDrawer(VAR);
    if (drawer != NULL)
    {
        VarDrawerSettings *settings = drawer->GetSettings();
        settings->VarMaskColorChanged(color);
        drawer->GetCanvas()->DeleteTexture(VAR_HASH, -1);
        drawer->Refresh();
    }
}

void ViewManager::VarMaskAlphaChanged(float alpha)
{
    VarDrawer *drawer = (VarDrawer*)GetDrawer(VAR);
    if (drawer != NULL)
    {
        VarDrawerSettings *settings = drawer->GetSettings();
        settings->VarMaskAlphaChanged(alpha);
        drawer->GetCanvas()->DeleteTexture(VAR_HASH, -1);
        drawer->Refresh();
    }
}

void ViewManager::VarMaskChanged()
{
    VarDrawer *drawer = (VarDrawer*)GetDrawer(VAR);
    if (drawer != NULL)
    {
        VarDrawerSettings *settings = drawer->GetSettings();
        drawer->GetCanvas()->DeleteTexture(VAR_HASH, -1);
        drawer->Refresh();
    }
}

void ViewManager::VarPositionReset()
{
    VarDrawer *drawer = (VarDrawer*)GetDrawer(VAR);
    if (drawer != NULL)
    {
        VarDrawerSettings *settings = drawer->GetSettings();
        settings->SetXYFromDataSet(true);
        drawer->Refresh();
    }
}

void ViewManager::GlyphDisplayModeReset(int i)
{
    GlyphDrawer *drawer = (GlyphDrawer*)GetDrawer(GLYPH);
    if (drawer != NULL)
    {
        GlyphDrawerSettings *settings = drawer->GetSettings();
        settings->SetXYFromDataSet(true);
        settings->SetDisplayMode(i);
        drawer->Refresh();
    }
}

void ViewManager::VarDisplayModeReset(int i)
{
    VarDrawer *drawer = (VarDrawer*)GetDrawer(VAR);
    if (drawer != NULL)
    {
        VarDrawerSettings *settings = drawer->GetSettings();
        settings->SetXYFromDataSet(true);
        settings->SetDisplayMode(i);
        drawer->Refresh();
    }
}

void ViewManager::VarLabelingModeReset(int i)
{
    VarDrawer *drawer = (VarDrawer*)GetDrawer(VAR);
    if (drawer != NULL)
    {
        if (i == 1)
            drawer->RemoveLayer(DIMENSIONNAME); 
        else
        {
            drawer->RemoveLayer(DIMENSIONNAME); 
            drawer->AddLayer(DIMENSIONNAME);
        }
        drawer->Refresh();
    }
}


void ViewManager::GlyphPositionReset()
{
    GlyphDrawer *drawer = (GlyphDrawer*)GetDrawer(GLYPH);
    if (drawer != NULL)
    {
        GlyphDrawerSettings *settings = drawer->GetSettings();
        settings->SetXYFromDataSet(true);
        drawer->Refresh();
    }
}

void ViewManager::FocusItemReset(int newID)
{
    GlyphDrawer *drawer = (GlyphDrawer*)GetDrawer(GLYPH);
    if (drawer != NULL)
    {
        GlyphDrawerSettings *settings = drawer->GetSettings();
        settings->BringTextureToFront(newID);
        settings->SetFocusTextureID(newID);
        drawer->Refresh();
    }
}

void ViewManager::BaseItemReset(int id)
{
    GlyphDrawer *drawer = (GlyphDrawer*)GetDrawer(GLYPH);
    if (drawer != NULL)
    {
        GlyphDrawerSettings *settings = drawer->GetSettings();
        settings->BringTextureToFront(id);
        settings->SetBaseTextureID(id);
    }
}

void ViewManager::BaseDimensionReset(int id)
{
    VarDrawer *drawer = (VarDrawer*)GetDrawer(VAR);
    if (drawer != NULL)
    {
        VarDrawerSettings *settings = drawer->GetSettings();
        settings->BringTextureToFront(id);
        settings->SetBaseTextureID(id);
    }
}

void ViewManager::RefreshGlyphView()
{
    GlyphDrawer *drawer = (GlyphDrawer*)GetDrawer(GLYPH);
    if (drawer != NULL)
    {
        drawer->Refresh();
    }
}

void ViewManager::SelectedItemsChanged()
{
    GlyphDrawer *drawer = (GlyphDrawer*)GetDrawer(GLYPH);
    if (drawer != NULL)
    {
        drawer->SelectedItemsChanged();
        drawer->Refresh();
    }

    VarDrawer *drawer1 = (VarDrawer*)GetDrawer(VAR);
    if (drawer1 != NULL)
    {
        VarDrawerSettings *settings = drawer1->GetSettings();
        settings->SelectedItemsChanged();
        drawer1->SelectedItemsChanged();
        drawer1->Refresh();
    }

}

void ViewManager::VarTextureChanged()
{
    VarDrawer *drawer1 = (VarDrawer*)GetDrawer(VAR);
    if (drawer1 != NULL)
    {
        drawer1->GetCanvas()->DeleteTextures(VAR_HASH);

        drawer1->Refresh();
    }

}

void ViewManager::TimeHistogramTextureChanged()
{
    VarDrawer *drawer1 = (VarDrawer*)GetDrawer(VAR);
    if (drawer1 != NULL)
    {
        drawer1->GetCanvas()->DeleteTextures(TIME_HISTOGRAM_HASH);

        drawer1->Refresh();
    }

}

void ViewManager::SetVarGlyphMode(int i)
{
    VarDrawer *drawer1 = (VarDrawer*)GetDrawer(VAR);
    if (drawer1 != NULL)
    {
        if (i != 4 && i != 5)
        {
            drawer1->RemoveLayer(COLUMN);
            drawer1->RemoveLayer(TIMEHISTOGRAM);
            drawer1->RemoveLayer(DIMENSIONNAME);
            drawer1->AddLayer(COLUMN);
            drawer1->AddLayer(DIMENSIONNAME);

            if (i == 0)
                drawer1->GetSettings()->SetGlyphStyle(PIXEL_SPIRAL);
            else if (i == 1)
                drawer1->GetSettings()->SetGlyphStyle(PIXEL_VERTICAL);
            else if (i == 2)
                drawer1->GetSettings()->SetGlyphStyle(PIXEL_HORI);

            drawer1->GetCanvas()->DeleteTextures(VAR_HASH);
            drawer1->Refresh();
        }
        else
        {
            drawer1->RemoveLayer(TIMEHISTOGRAM);
            drawer1->RemoveLayer(COLUMN);
            drawer1->RemoveLayer(DIMENSIONNAME);
            drawer1->AddLayer(TIMEHISTOGRAM);
            drawer1->AddLayer(DIMENSIONNAME);
            drawer1->GetCanvas()->DeleteTextures(TIME_HISTOGRAM_HASH);
            drawer1->Refresh();
        }
    }
}

void ViewManager::SelectedDimensionsChanged()
{
/*    GlyphDrawer *drawer = (GlyphDrawer*)GetDrawer(GLYPH);
    if (drawer != NULL)
    {
        drawer->SelectedItemsChanged();
        drawer->Refresh();
    }
*/
    VarDrawer *drawer1 = (VarDrawer*)GetDrawer(VAR);
    if (drawer1 != NULL)
    {
        VarDrawerSettings *settings = drawer1->GetSettings();
        settings->SelectedItemsChanged();
//        settings->SetShowSelectedOnly(true);
//        docManager->SetUseSelectedDimensionSpace(true);

        drawer1->Refresh();
    }

}


void ViewManager::GlyphZPosRandomReset()
{
    GlyphDrawer *drawer = (GlyphDrawer*)GetDrawer(GLYPH);
    if (drawer != NULL)
    {
        GlyphDrawerSettings *settings = drawer->GetSettings();
        settings->SetZPosStyle(ZPOS_RANDOM);
        settings->ResetZPos();
        drawer->Refresh();
    }
}

void ViewManager::VarZPosRandomReset()
{
    VarDrawer *drawer = (VarDrawer*)GetDrawer(VAR);
    if (drawer != NULL)
    {
        VarDrawerSettings *settings = drawer->GetSettings();
        settings->SetZPosStyle(ZPOS_RANDOM);
        settings->ResetZPos();
        drawer->Refresh();
    }
}

